#!/usr/bin/env python
"""
_WMTweak_

Define extraction of a standard set of WM related PSet parameters
Note: This can be used within the CMSSW environment to act on a
process/config but does not depend on any CMSSW libraries. It needs to stay like this.

"""
from __future__ import print_function, division

from builtins import map, range, str, object
from future.utils import viewitems, viewkeys

import logging
import os
import pickle
from Utils.PythonVersion import PY3
from Utils.Utilities import encodeUnicodeToBytesConditional
from PSetTweaks.PSetTweak import PSetTweak

# params to be extracted from an output module
_TweakOutputModules = [
    "fileName",
    "logicalFileName",
    "compressionLevel",
    "basketSize",
    "splitLevel",
    "overrideInputFileSplitLevels",
    "maxSize",
    "fastCloning",
    "sortBaskets",
    "dropMetaData",
    # "outputCommands", #this is just a huge pile of stuff which we probably shouldnt be setting anyways
    "SelectEvents.SelectEvents",
    "dataset.dataTier",
    "dataset.filterName",
    # TODO: support dataset.* here

]

_TweakParams = [
    # options
    "process.options.fileMode",
    "process.options.wantSummary",
    "process.options.allowUnscheduled",
    "process.options.makeTriggerResults",
    "process.options.Rethrow",
    "process.options.SkipEvent",
    "process.options.FailPath",
    "process.options.FailModule",
    "process.options.IgnoreCompletely",

    # config metadata
    "process.configurationMetadata.name",
    "process.configurationMetadata.version",
    "process.configurationMetadata.annotation",

    # source
    "process.source.maxEvents",
    "process.source.skipEvents",
    "process.source.firstEvent",
    "process.source.firstRun",
    "process.source.firstLuminosityBlock",
    "process.source.numberEventsInRun",
    "process.source.fileNames",
    "process.source.secondaryFileNames",
    "process.source.fileMatchMode",
    "process.source.overrideCatalog",
    "process.source.numberEventsInLuminosityBlock",
    "process.source.firstTime",
    "process.source.timeBetweenEvents",
    "process.source.eventCreationDelay",
    "process.source.needSecondaryFileNames",
    "process.source.parametersMustMatch",
    "process.source.branchesMustMatch",
    "process.source.setRunNumber",
    "process.source.skipBadFiles",
    "process.source.eventsToSkip",
    "process.source.lumisToSkip",
    "process.source.eventsToProcess",
    "process.source.lumisToProcess",
    "process.source.noEventSort",
    "process.source.duplicateCheckMode",
    "process.source.inputCommands",
    "process.source.dropDescendantsOfDroppedBranches",

    # maxevents
    "process.maxEvents.input",
    "process.maxEvents.output",
    # TODO: there are more settings stored as a VPSet which are a complete
    #       ballache to handle, suggest asking framework to change interface here

    # job report service
    # Everything has shifted to the default cff

    # message logger
    # Everything is in the default cff

    # random seeds
    "process.RandomNumberGeneratorService.*.initialSeed",
    "process.GlobalTag.globaltag",

]


class WMTweakMaskError(Exception):
    def __init__(self, mask=None, msg="Cannot set process from job mask"):
        super(WMTweakMaskError, self).__init__()
        self.mask = mask
        self.message = msg

    def __str__(self):
        return "Error: %s \n Mask: %s" % (self.message, str(self.mask))


def lfnGroup(job):
    """
    _lfnGroup_

    Determine the lfnGroup from the job counter and the agent number
    provided in the job baggage, the job counter and agent number
    default both to 0. The result will be a 5-digit string.
    """
    modifier = str(job.get("agentNumber", 0))
    jobLfnGroup = modifier + str(job.get("counter", 0) // 1000).zfill(4)
    return jobLfnGroup


def hasParameter(pset, param, nopop=False):
    """
    _hasParameter_

    check that pset provided has the attribute chain
    specified.

    Eg if param is pset.attr1.attr2.attr3
    check for pset.attr1.attr2.attr3
    returns True if parameter exists, False if not

    """
    params = param.split(".")
    if not nopop:
        params.pop(0)  # first param is the pset we have the reference to
    lastParam = pset
    for param in params:
        lastParam = getattr(lastParam, param, None)
        if lastParam is None:
            return False
    if lastParam is not None:
        return True
    return False


def getParameter(pset, param, nopop=False):
    """
    _getParameter_

    Retrieve the specified parameter from the PSet Provided
    given the attribute chain

    returns None if not found

    """
    params = param.split(".")
    if not nopop:
        params.pop(0)  # first param is the pset we have the reference to
    lastParam = pset
    for param in params:
        lastParam = getattr(lastParam, param, None)
        if lastParam is None:
            return None
    return lastParam.value()


def setParameter(process, param, value):
    """
    _setParameter_

    Set the value of the parameter to the given value.

    - process is the reference to the process

    - param is the name of the param as process.pset1.pset2...parameter

    - value is the value to set that paramter to

    """
    params = param.split('.')
    params.pop(0)  # first is process object
    lastPSet = process
    for pset in params:
        lastPSet = getattr(lastPSet, pset, None)
        if lastPSet is None:
            msg = "Cannot find attribute named: %s\n" % pset
            msg += "Cannot set value: %s" % param
            logging.error(msg)
            return

    lastPSet.setValue(value)
    return


def expandParameter(process, param):
    """
    _expandParameter_

    If param contains a wildcard * then expand it to the list of
    matching parameters

    """
    params = param.split('.')
    params.pop(0)
    lastResults = {"process": process}
    finalResults = {}
    for _ in range(0, len(params)):
        pset = params.pop(0)
        if pset == "*":
            newResults = {}
            for lastResultKey, lastResultVal in viewitems(lastResults):
                for param in listParams(lastResultVal):
                    newResultKey = "%s.%s" % (lastResultKey, param)
                    newResultVal = getattr(lastResultVal, param)
                    if not hasattr(newResultVal, "parameters_"):
                        if len(params) == 0:
                            finalResults[newResultKey] = newResultVal
                        continue
                    newResults[newResultKey] = newResultVal
            lastResults = newResults


        else:
            newResults = {}
            for lastResultKey, lastResultVal in viewitems(lastResults):
                newResultKey = "%s.%s" % (lastResultKey, pset)
                newResultVal = getattr(lastResultVal, pset, None)
                if not hasattr(newResultVal, "parameters_"):
                    finalResults[newResultKey] = newResultVal
                    continue
                newResults[newResultKey] = newResultVal

            lastResults = newResults

    return finalResults


listParams = lambda x: [y for y in x.parameters_()]


class TweakMaker(object):
    """
    _TweakMaker_

    Object to generate a Tweak instance from a generic
    configuration by searching for a set of specific parameters
    within the process, all output modules and a set of parameters
    within the output modules

    """

    def __init__(self, processParams=None, outmodParams=None):
        processParams = processParams or _TweakParams
        outmodParams = outmodParams or _TweakOutputModules
        self.processLevel = processParams
        self.outModLevel = outmodParams

    def __call__(self, process):
        tweak = PSetTweak()
        # handle process parameters
        processParams = []
        for param in self.processLevel:
            processParams.extend(viewkeys(expandParameter(process, param)))

        for param in processParams:
            if hasParameter(process, param):
                tweak.addParameter(param, getParameter(process, param))

        # output modules
        tweak.addParameter('process.outputModules_', [])
        for outMod in process.outputModules_():
            tweak.getParameter('process.outputModules_').append(outMod)
            outModRef = getattr(process, outMod)
            for param in self.outModLevel:
                fullParam = "process.%s.%s" % (outMod, param)
                if hasParameter(outModRef, param, True):
                    tweak.addParameter(fullParam, getParameter(outModRef, param, True))

        return tweak


def makeTweak(process):
    """
    _makeTweak_

    Create a PSetTweak instance using the list of potential parameters
    defined above. If the process has those parameters, they get added
    to the tweak, if not, they are left out.

    """
    maker = TweakMaker()
    return maker(process)


def applyTweak(process, tweak, fixup=None):
    """
    _applyTweak_

    Add the changes contained in the tweak to the process to give a job specific
    process.  The fixup parameters is a dictionary keyed by parameter name.  If
    the tweak contains a parameter in the dictionary the value in the dict will
    be calls and passed the process.

    This is useful for preparing the process before the value is applied (ie-
    making sure all the necessary PSets and configuration values exist).
    """
    for param, value in tweak:
        if isinstance(value, type(u'')) and hasattr(value, "encode"):
            logging.info("Found unicode parameter type for param: %s, with value: %s", param, value)
            value = value.encode("utf-8")
        if fixup and param in fixup:
            fixup[param](process)

        setParameter(process, param, value)


childParameters = lambda p, x: [i for i in x._internal_settings if i not in x._internal_children]
childSections = lambda s: [getattr(s, x) for x in s._internal_children]


class ConfigSectionDecomposer(object):
    """
    _ConfigSectionDecomposer_

    Util to collapse a ConfigSection to a dict of . delimited param: values
    where the params contain the section structure.

    May turn out to be generally useful for ConfigSections

    """

    def __init__(self):
        self.configSects = []
        self.parameters = {}
        self.queue = []

    def __call__(self, configSect):
        """
        _operator(configSect)_

        recursively traverse all parameters in this and all child
        PSets

        """
        self.queue.append(configSect._internal_name)
        csectPath = ".".join(self.queue)
        self.configSects.append(csectPath)
        params = childParameters(csectPath, configSect)
        for par in params:
            paramName = ".".join([csectPath, par])
            paramVal = getattr(configSect, par)
            self.parameters[paramName] = paramVal

        list(map(self, childSections(configSect)))
        self.queue.pop(-1)


def decomposeConfigSection(csect):
    """
    _decomposeConfigSection_

    Util to convert a config section into a . delimited dict of
    parameters mapped to values

    """
    decomposer = ConfigSectionDecomposer()
    decomposer(csect)

    return decomposer.parameters


def makeTaskTweak(stepSection, result):
    """
    _makeTaskTweak_

    Create a tweak for options in the task that apply to all jobs.
    """

    # GlobalTag
    if hasattr(stepSection, "application"):
        if hasattr(stepSection.application, "configuration"):
            if hasattr(stepSection.application.configuration, "pickledarguments"):
                pklArgs = encodeUnicodeToBytesConditional(stepSection.application.configuration.pickledarguments,
                                                          condition=PY3)
                args = pickle.loads(pklArgs)
                if 'globalTag' in args:
                    result.addParameter("process.GlobalTag.globaltag",  "customTypeCms.string('%s')" % args['globalTag'])

    return


def makeJobTweak(job, result):
    """
    _makeJobTweak_

    Convert information from a WMBS Job object into a PSetTweak
    that can be used to modify a CMSSW process.
    """
    baggage = job.getBaggage()

    # Check in the baggage if we are processing .lhe files
    lheInput = getattr(baggage, "lheInputFiles", False)

    # Input files and secondary input files.
    primaryFiles = []
    secondaryFiles = []
    for inputFile in job["input_files"]:
        if inputFile["lfn"].startswith("MCFakeFile"):
            # If there is a preset lumi in the mask, use it as the first
            # luminosity setting
            if job['mask'].get('FirstLumi', None) != None:
                logging.info("Setting 'firstLuminosityBlock' attr to: %s", job['mask']['FirstLumi'])
                result.addParameter("process.source.firstLuminosityBlock",
                                    "customTypeCms.untracked.uint32(%s)" % job['mask']['FirstLumi'])
            else:
                # We don't have lumi information in the mask, raise an exception
                raise WMTweakMaskError(job['mask'],
                                       "No first lumi information provided")
            continue

        primaryFiles.append(inputFile["lfn"])
        for secondaryFile in inputFile["parents"]:
            secondaryFiles.append(secondaryFile["lfn"])

    logging.info("Adding %d files to 'fileNames' attr", len(primaryFiles))
    logging.info("Adding %d files to 'secondaryFileNames' attr", len(secondaryFiles))
    if len(primaryFiles) > 0:
        result.addParameter("process.source.fileNames", "customTypeCms.untracked.vstring(%s)" % primaryFiles)
        if len(secondaryFiles) > 0:
            result.addParameter("process.source.secondaryFileNames", "customTypeCms.untracked.vstring(%s)" % secondaryFiles)
    elif not lheInput:
        # First event parameter should be set from whatever the mask says,
        # That should have the added protection of not going over 2^32 - 1
        # If there is nothing in the mask, then we fallback to the counter method
        if job['mask'].get('FirstEvent', None) != None:
            logging.info("Setting 'firstEvent' attr to: %s", job['mask']['FirstEvent'])
            result.addParameter("process.source.firstEvent", "customTypeCms.untracked.uint32(%s)" % job['mask']['FirstEvent'])
        else:
            # No first event information in the mask, raise and error
            raise WMTweakMaskError(job['mask'],
                                   "No first event information provided in the mask")

    mask = job['mask']

    # event limits
    maxEvents = mask.getMaxEvents()
    if maxEvents is None:
        maxEvents = -1
    logging.info("Setting 'maxEvents.input' attr to: %s", maxEvents)
    result.addParameter("process.maxEvents", "customTypeCms.untracked.PSet(input=cms.untracked.int32(%s))"% maxEvents)

    # We don't want to set skip events for MonteCarlo jobs which have
    # no input files.
    firstEvent = mask['FirstEvent']
    if firstEvent != None and firstEvent >= 0 and (len(primaryFiles) > 0 or lheInput):
        if lheInput:
            logging.info("Setting 'skipEvents' attr to: %s", firstEvent - 1)
            result.addParameter("process.source.skipEvents", "customTypeCms.untracked.uint32(%s)" % (firstEvent - 1))
        else:
            logging.info("Setting 'skipEvents' attr to: %s", firstEvent)
            result.addParameter("process.source.skipEvents", "customTypeCms.untracked.uint32(%s)" % firstEvent)

    firstRun = mask['FirstRun']
    if firstRun != None:
        result.addParameter("process.source.firstRun", "customTypeCms.untracked.uint32(%s)" % firstRun)
    elif not len(primaryFiles):
        # Then we have a MC job, we need to set firstRun to 1
        logging.debug("MCFakeFile initiated without job FirstRun - using one.")
        result.addParameter("process.source.firstRun", "customTypeCms.untracked.uint32(1)")

    runs = mask.getRunAndLumis()
    lumisToProcess = []
    for run in viewkeys(runs):
        lumiPairs = runs[run]
        for lumiPair in lumiPairs:
            if len(lumiPair) != 2:
                # Do nothing
                continue
            lumisToProcess.append("%s:%s-%s:%s" % (run, lumiPair[0], run, lumiPair[1]))

    if len(lumisToProcess) > 0:
        logging.info("Adding %d run/lumis mask to 'lumisToProcess' attr", len(lumisToProcess))
        result.addParameter("process.source.lumisToProcess", "customTypeCms.untracked.VLuminosityBlockRange(%s)" % lumisToProcess)

    # install any settings from the per job baggage
    procSection = getattr(baggage, "process", None)
    if procSection is None:
        return result

    baggageParams = decomposeConfigSection(procSection)
    for k, v in viewitems(baggageParams):
        if isinstance(v, str):
            v =  "customTypeCms.untracked.string(%s)" % v
        elif isinstance(v, int):
            v =  "customTypeCms.untracked.uint32(%s)" % v
        elif isinstance(v, list):
            v =  "customTypeCms.untracked.vstring(%s)" % v
        result.addParameter(k, v)

    return


def makeOutputTweak(outMod, job, result):
    """
    _makeOutputTweak_

    Make a PSetTweak for the output module and job instance provided

    """
    # output filenames
    modName = outMod.getInternalName()
    logging.info("modName = %s", modName)
    fileName = "%s.root" % modName

    result.addParameter("process.%s.fileName" % modName, fileName)

    lfnBase = str(getattr(outMod, "lfnBase", None))
    if lfnBase != None:
        lfn = "%s/%s/%s.root" % (lfnBase, lfnGroup(job), modName)
        result.addParameter("process.%s.logicalFileName" % modName, lfn)

    return


def readAdValues(attrs, adname, castInt=False):
    """
    A very simple parser for the ads available at runtime.  Returns
    a dictionary containing
    - attrs: A list of string keys to look for.
    - adname: Which ad to parse; "job" for the $_CONDOR_JOB_AD or
      "machine" for $_CONDOR_MACHINE_AD
    - castInt: Set to True to force the values to be integer literals.
      Otherwise, this will return the values as a string representation
      of the ClassAd expression.

    Note this is not a ClassAd parser - will not handle new-style ads
    or any expressions.

    Will return a dictionary containing the key/value pairs that were
    present in the ad and parseable.

    On error, returns an empty dictionary.
    """
    retval = {}
    adfile = None
    if adname == 'job':
        adfile = os.environ.get("_CONDOR_JOB_AD")
    elif adname == 'machine':
        adfile = os.environ.get("_CONDOR_MACHINE_AD")
    else:
        logging.warning("Invalid ad name requested for parsing: %s", adname)
        return retval
    if not adfile:
        logging.warning("%s adfile is not set in environment.", adname)
        return retval
    attrs = [i.lower() for i in attrs]

    try:
        with open(adfile) as fd:
            for line in fd:
                info = line.strip().split("=", 1)
                if len(info) != 2:
                    continue
                attr = info[0].strip().lower()
                if attr in attrs:
                    val = info[1].strip()
                    if castInt:
                        try:
                            retval[attr] = int(val)
                        except ValueError:
                            logging.warning("Error parsing %s's %s value: %s", adname, attr, val)
                    else:
                        retval[attr] = val
    except IOError:
        logging.exception("Error opening %s ad:", adname)
        return {}

    return retval


def resizeResources(resources):
    """
    _resizeResources_

    Look at the job runtime environment and determine whether we are allowed
    to resize the core count.  If so, change the resources dictionary passed
    to this function according to the information found in $_CONDOR_MACHINE_AD.
    The following keys are changed:
     - cores -> uses value of Cpus from the machine ad.
     - memory -> Memory

    This only works when running under HTCondor, $_CONDOR_MACHINE_AD exists,
    and WMCore_ResizeJob is true.
          - WMCore_ResizeJob is 'true'

    No return value - the resources directory is changed in-place.
    Should not throw an exception - on error, no change is made and a message
    is printed out.
    """
    if readAdValues(['wmcore_resizejob'], 'job').get('wmcore_resizejob', 'false').lower() != "true":
        logging.info("Not resizing job")
        return

    logging.info("Resizing job.  Initial resources: %s", resources)
    adValues = readAdValues(['memory', 'cpus'], 'machine', castInt=True)
    machineCpus = adValues.get('cpus', 0)
    machineMemory = adValues.get('memory', 0)
    if machineCpus > 0 and 'cores' in resources:
        resources['cores'] = machineCpus
    if machineMemory > 0 and 'memory' in resources:
        resources['memory'] = machineMemory
    logging.info("Resizing job.  Resulting resources: %s", resources)



#!/usr/bin/env python
"""
_JobAccountant_

JobAccountant harness.  Instantiate the JobAccountandPoller and have it poll
WMBS for complete jobs.
"""




import threading

from WMCore.Agent.Harness import Harness
from WMComponent.JobAccountant.JobAccountantPoller import JobAccountantPoller

class JobAccountant(Harness):
    def __init__(self, config):
        Harness.__init__(self, config)
        return

    def preInitialization(self):
        pollInterval = self.config.JobAccountant.JobAccountantPoller.pollInterval
        myThread = threading.currentThread()
        myThread.workerThreadManager.addWorker(JobAccountantPoller(self.config), pollInterval)



#!/usr/bin/env python
"""
The actual taskArchiver algorithm

Procedure:
a) Find and marks a finished all newly finished subscriptions
      This is defined by the Subscriptions.MarkNewFinishedSubscriptions DAO
b) Look for finished workflows as defined in the Workflow.GetFinishedWorkflows DAO
c) Upload couch summary information
d) Call WMBS.Subscription.deleteEverything() on all the associated subscriptions
e) Delete couch information and working directories

This should be a simple process.  Because of the long time between
the submission of subscriptions projected and the short time to run
this class, it should be run irregularly.


Config options
histogramKeys: Allows you to report values in histogram form in the
  workloadSummary - i.e., as a list of bins
histogramBins: Bin size for all histograms
histogramLimit: Limit in terms of number of standard deviations from the
  average at which you cut the histogram off.  All points outside of that
  go into overflow and underflow.
"""
from builtins import str as newstr, bytes
from future.utils import viewvalues

import logging
import threading
import traceback

from Utils.Timers import timeFunction
from Utils.wmcoreDTools import resetWatchdogTimer, moduleName
from WMComponent.TaskArchiver.DataCache import DataCache
from WMCore.DAOFactory import DAOFactory
from WMCore.Services.ReqMgr.ReqMgr import ReqMgr
from WMCore.Services.RequestDB.RequestDBWriter import RequestDBWriter
from WMCore.WMException import WMException
from WMCore.WorkQueue.WorkQueue import localQueue
from WMCore.WorkQueue.WorkQueueExceptions import WorkQueueNoMatchingElements
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread


class TaskArchiverPollerException(WMException):
    """
    _TaskArchiverPollerException_

    This is the class that serves as the customized
    Exception class for the TaskArchiverPoller

    As if you couldn't tell that already
    """
    pass


class TaskArchiverPoller(BaseWorkerThread):
    """
    Polls for Ended jobs

    List of attributes

    requireCouch:  raise an exception on couch failure instead of ignoring
    """

    def __init__(self, config):
        """
        Initialise class members
        """
        BaseWorkerThread.__init__(self)

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)

        self.dbsDaoFactory = DAOFactory(package="WMComponent.DBS3Buffer",
                                        logger=myThread.logger,
                                        dbinterface=myThread.dbi)

        self.config = config
        self.jobCacheDir = self.config.JobCreator.jobCacheDir

        if getattr(self.config.TaskArchiver, "useWorkQueue", False):
            # Get workqueue setup from config unless overridden
            if hasattr(self.config.TaskArchiver, 'WorkQueueParams'):
                self.workQueue = localQueue(**self.config.TaskArchiver.WorkQueueParams)
            else:
                from WMCore.WorkQueue.WorkQueueUtils import queueFromConfig
                self.workQueue = queueFromConfig(self.config)
        else:
            self.workQueue = None

        self.timeout = getattr(self.config.TaskArchiver, "timeOut", None)
        self.useReqMgrForCompletionCheck = getattr(self.config.TaskArchiver, 'useReqMgrForCompletionCheck', True)

        if not self.useReqMgrForCompletionCheck:
            # sets the local monitor summary couch db
            self.tier0CompletedState = "completed"
            self.requestLocalCouchDB = RequestDBWriter(self.config.AnalyticsDataCollector.localT0RequestDBURL,
                                                       couchapp=self.config.AnalyticsDataCollector.RequestCouchApp)
            self.centralCouchDBWriter = self.requestLocalCouchDB
        else:
            self.tier0CompletedState = None
            self.centralCouchDBWriter = RequestDBWriter(self.config.AnalyticsDataCollector.centralRequestDBURL)

            self.reqmgr2Svc = ReqMgr(self.config.General.ReqMgr2ServiceURL)

        # Load the cleanout state ID and save it
        stateIDDAO = self.daoFactory(classname="Jobs.GetStateID")
        self.stateID = stateIDDAO.execute("cleanout")

        return

    def terminate(self, params):
        """
        _terminate_

        This function terminates the job after a final pass
        """
        logging.debug("terminating. doing one more pass before we die")
        self.algorithm(params)
        return

    @timeFunction
    def algorithm(self, parameters=None):
        """
        _algorithm_

        Executes the two main methods of the poller:
        1. findAndMarkFinishedSubscriptions
        2. completeTasks
        Final result is that finished workflows get their summary built and uploaded to couch,
        and all traces of them are removed from the agent WMBS and couch (this last one on demand).
        """
        try:
            self.findAndMarkFinishedSubscriptions()
            (finishedwfs, finishedwfsWithLogCollectAndCleanUp) = self.getFinishedWorkflows()
            # set the data cache which can be used other thread (no ther thread should set the data cache)
            DataCache.setFinishedWorkflows(finishedwfsWithLogCollectAndCleanUp)
            self.completeTasks(finishedwfs)
        except WMException:
            myThread = threading.currentThread()
            if getattr(myThread, 'transaction', False) \
                    and getattr(myThread.transaction, 'transaction', False):
                myThread.transaction.rollback()
            raise
        except Exception as ex:
            myThread = threading.currentThread()
            msg = "Caught exception in TaskArchiver\n"
            msg += str(ex)
            if getattr(myThread, 'transaction', False) \
                    and getattr(myThread.transaction, 'transaction', False):
                myThread.transaction.rollback()
            raise TaskArchiverPollerException(msg)

        # Reset its own watchdog timer at the end of the run cycle
        logging.info(f"Resetting {moduleName(self)} watchdog timer.")
        if resetWatchdogTimer(self):
            logging.warning(f"Failed to reset {moduleName(self)} watchdog timer. The component might be restarted soon.")

        return

    def findAndMarkFinishedSubscriptions(self):
        """
        _findAndMarkFinishedSubscriptions_

        Find new finished subscriptions and mark as finished in WMBS.
        """
        myThread = threading.currentThread()

        myThread.transaction.begin()

        # Get the subscriptions that are now finished and mark them as such
        logging.info("Polling for finished subscriptions")
        finishedSubscriptions = self.daoFactory(classname="Subscriptions.MarkNewFinishedSubscriptions")
        finishedSubscriptions.execute(self.stateID, timeOut=self.timeout)
        logging.info("Finished subscriptions updated")

        myThread.transaction.commit()

        return

    def getFinishedWorkflows(self):
        """
        1. Get finished workflows (a finished workflow is defined in Workflow.GetFinishedWorkflows)
        2. Get finished workflows with logCollect and Cleanup only.
        3. combined those and make return
           finishedwfs - without LogCollect and CleanUp task
           finishedwfsWithLogCollectAndCleanUp - including LogCollect and CleanUp task
        """

        finishedWorkflowsDAO = self.daoFactory(classname="Workflow.GetFinishedWorkflows")
        finishedwfs = finishedWorkflowsDAO.execute()
        finishedLogCollectAndCleanUpwfs = finishedWorkflowsDAO.execute(onlySecondary=True)
        finishedwfsWithLogCollectAndCleanUp = {}
        for wf in finishedLogCollectAndCleanUpwfs:
            if wf in finishedwfs:
                finishedwfsWithLogCollectAndCleanUp[wf] = finishedwfs[wf]
        return (finishedwfs, finishedwfsWithLogCollectAndCleanUp)

    def killCondorJobsByWFStatus(self, statusList):
        if isinstance(statusList, (newstr, bytes)):
            statusList = [statusList]
        reqNames = self.centralCouchDBWriter.getRequestByStatus(statusList)
        logging.info("There are %d requests in %s status in central couch.", len(reqNames), statusList)
        if self.workQueue is not None:
            self.workQueue.killWMBSWorkflows(reqNames)
        return reqNames

    def completeTasks(self, finishedwfs):
        """
        _completeTasks_

        This method will call several auxiliary methods to do the following:

        1. Notify the WorkQueue about finished subscriptions
        2. mark workflow as completed in the dbsbuffer_workflow table
        """
        if not finishedwfs:
            return

        logging.info("Found %d candidate workflows for completing:", len(finishedwfs))
        completedWorkflowsDAO = self.dbsDaoFactory(classname="UpdateWorkflowsToCompleted")

        centralCouchAlive = True
        try:
            self.killCondorJobsByWFStatus(["force-complete", "aborted"])
        except Exception as ex:
            centralCouchAlive = False
            logging.error("we will try again when remote couch server comes back\n%s", str(ex))

        if centralCouchAlive:
            logging.info("Marking subscriptions as Done ...")
            for workflow in finishedwfs:
                try:
                    # Notify the WorkQueue, if there is one
                    if self.workQueue is not None:
                        subList = []
                        for l in viewvalues(finishedwfs[workflow]["workflows"]):
                            subList.extend(l)
                        self.notifyWorkQueue(subList)

                    # Tier-0 case, the agent has to mark it completed
                    if not self.useReqMgrForCompletionCheck:
                        resp = self.requestLocalCouchDB.updateRequestStatus(workflow, self.tier0CompletedState)
                        logging.info("Workflow %s updated to status '%s'. Response: %s",
                                     workflow, self.tier0CompletedState, resp)

                    completedWorkflowsDAO.execute([workflow])

                except TaskArchiverPollerException as ex:
                    # Something didn't go well when notifying the workqueue, abort!!!
                    logging.error("Something bad happened while archiving tasks.")
                    logging.error(str(ex))
                    continue
                except Exception as ex:
                    # Something didn't go well on couch, abort!!!
                    msg = "Problem while archiving tasks for workflow %s\n" % workflow
                    msg += "Exception message: %s" % str(ex)
                    msg += "\nTraceback: %s" % traceback.format_exc()
                    logging.error(msg)
                    continue
        return

    def notifyWorkQueue(self, subList):
        """
        _notifyWorkQueue_

        Tells the workQueue component that a particular subscription,
        or set of subscriptions, is done.  Receives confirmation
        """

        for sub in subList:
            try:
                self.workQueue.doneWork(SubscriptionId=sub)
            except WorkQueueNoMatchingElements:
                # Subscription wasn't known to WorkQueue, feel free to clean up
                logging.debug("Local WorkQueue knows nothing about this subscription: %s", sub)
            except Exception as ex:
                msg = "Error talking to workqueue: %s\n" % str(ex)
                msg += "Tried to complete the following: %s\n" % sub
                raise TaskArchiverPollerException(msg)

        return



#!/usr/bin/env python
"""
_Service_

A Service talks to some http(s) accessible service that provides information and
caches the result of these queries. The cache will be refreshed if the file is
older than a timeout set in the instance of Service.

It has a cache path, cache duration, an endpoint (the url the
service exists on) a logger and an accept type (json, xml etc) and method
(GET/POST).

The Service satisfies two caching cases:

1. set a defined query, cache results, poll for new ones
2. use a changing query, cache results to a file depending on the query, poll
   for new ones

Data maybe passed to the remote service either via adding the query string to
the URL (for GET's) or by passing a dictionary to either the service constructor
(case 1.) or by passing the data as a dictionary to the refreshCache,
forceCache, clearCache calls. By default the cache lasts 30 minutes.

Calling refreshCache/forceRefresh will return an open file object, the cache
file. Once done with it you should close the object.

The service has a default timeout to receive a response from the remote service
of 300 seconds. Over ride this by passing in a timeout via the configuration
dict, set to None if you want to turn off the timeout.

If you just want to retrieve the data without caching use the Requests class
directly.

The Service class provides two layers of caching:
    1. Caching from httplib2 is provided via Request, this respects etag and
    expires, but the cache will be lost if the service raises an exception or
    similar.
    2. Internal caching which respects an internal cache duration. If the remote
    service fails to respond the second layer cache will be used until the cache
    dies.

In tabular form:

httplib2 cache  |   yes    |   yes    |    no    |     no     |
----------------+----------+----------+----------+------------+
service cache   |    no    |   yes    |   yes    |     no     |
----------------+----------+----------+----------+------------+
result          |  cached  |  cached  |  cached  | not cached |
"""

from builtins import str
from future import standard_library
standard_library.install_aliases()

import datetime
import json
import logging
import os
import time
from io import BytesIO, StringIO
from http.client import HTTPException

from Utils.PythonVersion import PY3
from WMCore.Services.Requests import Requests, JSONRequests
from WMCore.WMException import WMException

try:
    from httplib2 import HttpLib2Error
except ImportError:
    # Mock HttpLib2Error since we don't want that WMCore depend on httplib2 using pycurl
    class HttpLib2Error(Exception):
        pass


def isfile(obj):
    """
    Check whether obj is a file-like object (file, StringIO)"
    """
    return hasattr(obj, 'flush')


def cache_expired(cache, delta=0):
    """
    Is the cache expired? At delta hours (default 0) in the future.
    """
    if isfile(cache):
        # already opened file-like object, assume we need to refresh
        return True
    else:
        # then it's a file name
        if not os.path.exists(cache):
            return True

        delta = datetime.timedelta(hours=delta)
        t = datetime.datetime.now() - delta
        # cache file mtime has been set to cache expiry time
        if os.path.getmtime(cache) < time.mktime(t.timetuple()):
            return True

    return False


def _makeHash(inputdata):
    """
    Turn the input data into json and hash the string. This is simple and
    means that the input data must be json-serialisable, which is good.
    """
    json_hash = json.dumps(inputdata)
    return json_hash.__hash__()


class Service(dict):
    def __init__(self, cfg_dict=None):
        super(Service, self).__init__()
        cfg_dict = cfg_dict or {}
        # The following should read the configuration class
        for a in ['endpoint']:
            assert a in list(cfg_dict), "Can't have a service without a %s" % a

        # if end point ends without '/', add that
        if not cfg_dict['endpoint'].endswith('/'):
            cfg_dict['endpoint'] = cfg_dict['endpoint'].strip() + '/'

        # set up defaults
        self.setdefault("inputdata", {})
        self.setdefault("cacheduration", 0.5)
        self.supportVerbList = ('GET', 'POST', 'PUT', 'DELETE')
        # this value should be only set when whole service class uses
        # the same verb ('GET', 'POST', 'PUT', 'DELETE')
        self.setdefault("method", None)

        # Set a timeout for the socket
        self.setdefault("timeout", 300)

        # then update with the incoming dict
        self.update(cfg_dict)

        self['service_name'] = self.__class__.__name__  # used for cache naming

        # Get the request class, to instantiate later
        # either passed as param to __init__, determine via scheme or default
        if isinstance(self.get('requests'), type):
            requests = self['requests']
        elif self.get('accept_type') == "application/json" and self.get('content_type') == "application/json":
            requests = JSONRequests
        else:
            requests = Requests
        # Instantiate a Request
        try:
            self["requests"] = requests(cfg_dict['endpoint'], cfg_dict)
        except WMException as ex:
            msg = str(ex)
            self["logger"].exception(msg)
            raise

        # cachepath will be modified - i.e. hostname added
        self['cachepath'] = self["requests"]["cachepath"]

        if 'logger' not in self:
            if self['cachepath']:
                logfile = os.path.join(self['cachepath'], '%s.log' % self.__class__.__name__.lower())
            else:
                logfile = None
            logging.basicConfig(level=logging.DEBUG,
                                format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                                datefmt='%m-%d %H:%M',
                                filename=logfile,
                                filemode='w')
            self['logger'] = logging.getLogger(self.__class__.__name__)
            self['requests']['logger'] = self['logger']

        msg = "Service '%s' initialized with the following settings: %s"
        self['logger'].debug(msg, self.__class__.__name__, self)

    def cacheFileName(self, cachefile, verb='GET', inputdata=None):
        """
        Calculate the cache filename for a given query.
        """
        # if not caching to disk return StringIO object
        if not self['cachepath'] or not cachefile:
            return StringIO() if PY3 else BytesIO()

        inputdata = inputdata or {}
        if inputdata:
            hash_ = _makeHash(inputdata)
        else:
            hash_ = _makeHash(self['inputdata'])
        cachefile = "%s/%s_%s_%s" % (self["cachepath"], hash_, verb, cachefile)

        return cachefile

    def refreshCache(self, cachefile, url='', inputdata=None, openfile=True,
                     encoder=True, decoder=True, verb='GET', contentType=None, incoming_headers=None, binary=False):
        """
        See if the cache has expired. If it has make a new request to the
        service for the input data. Return the cachefile as an open file object.
        If cachefile is None returns StringIO.

        :param cachefile: cache file name
        :param url: url value
        :param inputdata: HTTP (JSON) payload string
        :param openfile: flag to use open file, boolean
        :param encoder: flag to use encoder, boolean
        :param decoder: flag to use decode, boolean
        :param verb: HTTP method, string
        :param contentType: HTTP content type value, string
        :param incoming_headers: set of HTTP headers, dict
        :param binary: use binary mode to decode HTTP data, boolean
        :return: either file object or StringIO
        """
        inputdata = inputdata or {}
        incoming_headers = incoming_headers or {}
        verb = self._verbCheck(verb)

        cachefile = self.cacheFileName(cachefile, verb, inputdata)

        if cache_expired(cachefile, self["cacheduration"]):
            self.getData(cachefile, url, inputdata, incoming_headers, encoder, decoder, verb, contentType, binary=binary)
        else:
            self['logger'].debug('Data is from the Service cache')

        # cachefile may be filename or file object
        if openfile and not isfile(cachefile):
            if binary:
                return open(cachefile, 'rb')
            return open(cachefile, 'r', encoding='utf-8')
        else:
            return cachefile

    def forceRefresh(self, cachefile, url='', inputdata=None, openfile=True,
                     encoder=True, decoder=True, verb='GET',
                     contentType=None, incoming_headers=None, binary=False):
        """
        Make a new request to the service for the input data, regardless of the
        cache state. Return the cachefile as an open file object.
        If cachefile is None returns StringIO.

        :param cachefile: cache file name
        :param url: url value
        :param inputdata: HTTP (JSON) payload string
        :param openfile: flag to use open file, boolean
        :param encoder: flag to use encoder, boolean
        :param decoder: flag to use decode, boolean
        :param verb: HTTP method, string
        :param contentType: HTTP content type value, string
        :param incoming_headers: set of HTTP headers, dict
        :param binary: use binary mode to decode HTTP data, boolean
        :return: either file object or StringIO
        """
        inputdata = inputdata or {}
        incoming_headers = incoming_headers or {}
        verb = self._verbCheck(verb)

        cachefile = self.cacheFileName(cachefile, verb, inputdata)

        self['logger'].debug("Forcing cache refresh of %s" % cachefile)
        incoming_headers.update({'cache-control': 'no-cache'})
        self.getData(cachefile, url, inputdata, incoming_headers,
                     encoder, decoder, verb, contentType, force_refresh=True, binary=binary)
        if openfile and not isfile(cachefile):
            if binary:
                return open(cachefile, 'rb')
            return open(cachefile, 'r', encoding='utf-8')
        else:
            return cachefile

    def clearCache(self, cachefile, inputdata=None, verb='GET'):
        """
        Delete the cache file and the httplib2 cache.
        """
        if not self['cachepath'] or not cachefile:
            # nothing to clear
            return

        inputdata = inputdata or {}
        verb = self._verbCheck(verb)
        os.system("/bin/rm -f %s/*" % self['requests']['req_cache_path'])
        cachefile = self.cacheFileName(cachefile, verb, inputdata)
        try:
            if not isfile(cachefile):
                os.remove(cachefile)
        except OSError:  # File doesn't exist
            return

    def getData(self, cachefile, url, inputdata=None, incoming_headers=None,
                encoder=True, decoder=True,
                verb='GET', contentType=None, force_refresh=False, binary=False):
        """
        Takes the already generated *full* path to cachefile and the url of the
        resource. Don't need to call self.cacheFileName(cachefile, verb, inputdata)
        here.

        If cachefile is StringIO append to that

        :param cachefile: cache file name
        :param url: url value
        :param inputdata: HTTP (JSON) payload string
        :param encoder: flag to use encoder, boolean
        :param decoder: flag to use decode, boolean
        :param verb: HTTP method, string
        :param contentType: HTTP content type value, string
        :param force_refresh: refresh internal cache, boolean
        :param binary: use binary mode to decode HTTP data, boolean
        :return: None
        """
        inputdata = inputdata or {}
        incoming_headers = incoming_headers or {}
        verb = self._verbCheck(verb)

        try:
            # Get the data
            if not inputdata:
                inputdata = self["inputdata"]
            self['logger'].debug('getData: \n\turl: %s\n\tverb: %s\n\tincoming_headers: %s\n\tdata: %s',
                                 url, verb, incoming_headers, inputdata)
            # self['logger'].debug('getData: \n\turl: %s\n\tdata: %s' % \
            #                     (url, inputdata))
            data, dummyStatus, dummyReason, from_cache = self["requests"].makeRequest(uri=url,
                                                                                      verb=verb,
                                                                                      data=inputdata,
                                                                                      incoming_headers=incoming_headers,
                                                                                      encoder=encoder,
                                                                                      decoder=decoder,
                                                                                      contentType=contentType)
            if from_cache:
                # If it's coming from the cache we don't need to write it to the
                # second cache, or do we?
                self['logger'].debug('Data is from the Requests cache')
            else:
                # getData have done that for us
                if isfile(cachefile):
                    cachefile.write(data)
                    cachefile.seek(0, 0)  # return to beginning of file
                elif binary:
                    with open(cachefile, 'wb') as f:
                        f.write(data)
                elif isinstance(data, (dict, list)):
                    with open(cachefile, 'w', encoding='utf-8') as f:
                        f.write(json.dumps(data))
                else:
                    with open(cachefile, 'w', encoding='utf-8') as f:
                        f.write(data)


        except (IOError, HttpLib2Error, HTTPException) as he:
            #
            # Overly complicated exception handling. This is due to a request
            # from *Ops that it is very clear that data is is being returned
            # from a cachefile, and that cachefiles can be good/stale/dead.
            #
            if force_refresh or isfile(cachefile) or not os.path.exists(cachefile):
                msg = 'The cachefile %s does not exist and the service at %s'
                msg = msg % (cachefile, self["requests"]['host'] + url)
                if hasattr(he, 'status') and hasattr(he, 'reason'):
                    msg += ' is unavailable - it returned %s because %s' % (he.status,
                                                                            he.reason)
                    if hasattr(he, 'result'):
                        msg += ' with result: %s\n' % he.result
                else:
                    msg += ' raised a %s when accessed' % he.__repr__()
                self['logger'].warning(msg)
                raise he
            else:
                cache_dead = cache_expired(cachefile, delta=self["cacheduration"])
                if cache_dead:
                    msg = 'The cachefile %s is dead (older than %s hours of cache duration), '
                    msg += 'and the service at %s '
                    msg = msg % (cachefile, self["cacheduration"], url)
                    if hasattr(he, 'status') and hasattr(he, 'reason'):
                        msg += 'is unavailable - it returned %s because %s' % (he.status, he.reason)
                    else:
                        msg += 'raised a %s when accessed' % he.__repr__()
                    self['logger'].warning(msg)
                    raise he
                if self.get('usestalecache', False):
                    # then we can return data from the cache file, without raising an exception
                    # but with a suitable message in the log
                    msg = 'Returning stale cache data from %s, the service at ' % cachefile
                    if hasattr(he, 'status') and hasattr(he, 'reason'):
                        msg += '%s returned %s because %s' % (he.url, he.status, he.reason)
                    else:
                        msg += '%s raised a %s when accessed' % (url, he.__repr__())
                    self['logger'].warning(msg)
                else:
                    # Cache is not dead, but Service is configured to not return stale data.
                    msg = 'The cachefile %s is stale and the service at %s ' % (cachefile, url)
                    if hasattr(he, 'status') and hasattr(he, 'reason'):
                        msg += 'is unavailable - it returned %s because %s' % (he.status, he.reason)
                    else:
                        msg += 'raised a %s when accessed' % he.__repr__()
                    self['logger'].warning(msg)
                    raise he

    def _verbCheck(self, verb='GET'):
        if verb.upper() in self.supportVerbList:
            return verb.upper()
        elif self['method'].upper() in self.supportVerbList:
            return self['method'].upper()
        else:
            raise TypeError('verb parameter needs to be set')



#!/usr/bin/env python
"""
_ExistsForAccountant_

Oracle implementation of Files.ExistsForAccountant
"""

__all__ = []



from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.ExistsForAccountant import ExistsForAccountant as MySQLExistsForAccountant

class ExistsForAccountant(MySQLExistsForAccountant):
    """
    This is highly specialized.  You shouldn't confuse it with
    a normal Exists DAO
    """



#!/usr/bin/env python
"""
_FNALImpl_

Implementation of StageOutImpl interface for FNAL

"""
from __future__ import print_function

import os
import logging

from WMCore.Storage.Backends.LCGImpl import LCGImpl
from WMCore.Storage.Registry import registerStageOutImpl
from WMCore.Storage.StageOutImpl import StageOutImpl

_CheckExitCodeOption = True


def stripPrefixTOUNIX(filePath):
    if ".fnal.gov/" not in filePath:
        return filePath
    return filePath.split(".fnal.gov/", 1)[1]


class FNALImpl(StageOutImpl):
    """
    _FNALImpl_

    Implement interface for dcache xrootd command

    """

    def __init__(self, stagein=False):

        StageOutImpl.__init__(self, stagein)

        # Create and hold onto a srm implementation in case we need it
        self.srmImpl = LCGImpl(stagein)

    def storageMethod(self, PFN):
        """
        Return xrdcp or srm
        """

        method = 'local'  # default
        if PFN.startswith("root://"):
            method = 'xrdcp'
        if PFN.startswith("srm://"):
            method = 'srm'
        print("Using method %s for PFN %s" % (method, PFN))
        return method

    def createOutputDirectory(self, targetPFN):
        """
        _createOutputDirectory_

        Create a dir for the target pfn by translating it to
        a /dcache or /lustre name and calling mkdir
        we don't need to convert it, just mkdir.
        """
        method = self.storageMethod(targetPFN)

        if method == 'xrdcp':  # xrdcp autocreates parent directories
            return
        elif method == 'srm':
            self.srmImpl.createOutputDirectory(targetPFN)
        elif method == 'local':
            targetdir = os.path.dirname(targetPFN)
            command = "#!/bin/sh\n"
            command += "if [ ! -e \"%s\" ]; then\n" % targetdir
            command += " mkdir -p %s\n" % targetdir
            command += "fi\n"
            self.executeCommand(command)

    def createSourceName(self, protocol, pfn):
        """
        _createSourceName_

        generate the target PFN
        """
        method = self.storageMethod(pfn)

        if method == 'srm':
            return self.srmImpl.createSourceName(protocol, pfn)
        return pfn

    def createStageOutCommand(self, sourcePFN, targetPFN, options=None, checksums=None, authMethod=None, forceMethod=False):
        """
        _createStageOutCommand_

        Build a mkdir to generate the directory
        """
        logging.warning("Warning! FNALImpl does not support authMethod handling")

        if getattr(self, 'stageIn', False):
            return self.buildStageInCommand(sourcePFN, targetPFN, options)

        method = self.storageMethod(targetPFN)
        sourceMethod = self.storageMethod(sourcePFN)

        if ((method == 'srm' and sourceMethod == 'xrdcp') or
                (method == 'xrdcp' and sourceMethod == 'srm')):
            print("Incompatible methods for source and target")
            print("\tSource: method %s for PFN %s" % (sourceMethod, sourcePFN))
            print("\tTarget: method %s for PFN %s" % (method, targetPFN))
            return 1

        if method == 'srm' or sourceMethod == 'srm':
            return self.srmImpl.createStageOutCommand(sourcePFN, targetPFN, options)

        if method == 'xrdcp' or sourceMethod == 'xrdcp':
            original_size = os.stat(sourcePFN)[6]
            print("Local File Size is: %s" % original_size)

            useChecksum = (checksums != None and 'adler32' in checksums and not self.stageIn)
            if useChecksum:
                checksums['adler32'] = "%08x" % int(checksums['adler32'], 16)
                # non-functional in 3.3.1 xrootd clients due to bug
                # result += "-ODeos.targetsize=$LOCAL_SIZE\&eos.checksum=%s " % checksums['adler32']

                # therefor embed information into target URL
                targetPFN += "\?eos.targetsize=%s\&eos.checksum=%s" % (original_size, checksums['adler32'])
                print("Local File Checksum is: %s\"\n" % checksums['adler32'])

            # always overwrite the output

            result = "xrdcp-old -d 0 -f "
            if options != None:
                result += " %s " % options
            result += " %s " % sourcePFN
            result += " %s " % targetPFN
            result += """
            EXIT_STATUS=$?
            if [[ $EXIT_STATUS != 0 ]]; then
                echo "ERROR: xrdcp exited with $EXIT_STATUS"
            fi
            exit $EXIT_STATUS
            """
            return result

    def createDebuggingCommand(self, sourcePFN, targetPFN, options=None, checksums=None, authMethod=None, forceMethod=False):
        """
        Debug a failed fnal-flavor copy command for stageOut, without re-running it,
        providing information on the environment and the certifications

        :sourcePFN: str, PFN of the source file
        :targetPFN: str, destination PFN
        :options: str, additional options for copy command
        :checksums: dict, collect checksums according to the algorithms saved as keys
        """
        # Build the gfal-cp command for debugging purposes
        copyCommand = ""
        if getattr(self, 'stageIn', False):
            copyCommand += self.buildStageInCommand(sourcePFN, targetPFN, options)
        else:
            method = self.storageMethod(targetPFN)
            sourceMethod = self.storageMethod(sourcePFN)

            if ((method == 'srm' and sourceMethod == 'xrdcp') or (method == 'xrdcp' and sourceMethod == 'srm')):
                copyCommand += "Incompatible methods for source and target"
                copyCommand += "\tSource: method %s for PFN %s" % (sourceMethod, sourcePFN)
                copyCommand += "\tTarget: method %s for PFN %s" % (method, targetPFN)

            if method == 'srm' or sourceMethod == 'srm':
                copyCommand = self.srmImpl.createStageOutCommand(sourcePFN, targetPFN, options)

            if method == 'xrdcp' or sourceMethod == 'xrdcp':
                original_size = os.stat(sourcePFN)[6]
                copyCommand = "Local File Size is: %s" % original_size

                useChecksum = (checksums != None and 'adler32' in checksums and not self.stageIn)
                if useChecksum:
                    checksums['adler32'] = "%08x" % int(checksums['adler32'], 16)
                    targetPFN += "\?eos.targetsize=%s\&eos.checksum=%s" % (original_size, checksums['adler32'])
                    copyCommand += "Local File Checksum is: %s\"\n" % checksums['adler32']

                copyCommand += "xrdcp-old -d 0 -f "
                if options != None:
                    copyCommand += " %s " % options
                copyCommand += " %s " % sourcePFN
                copyCommand += " %s " % targetPFN
       
        result = self.debuggingTemplate.format(copy_command=copyCommand, source=sourcePFN, destination=targetPFN)
        return result

    def buildStageInCommand(self, sourcePFN, targetPFN, options=None):
        """
        _buildStageInCommand_

        Create normal xrdcp commad for staging in files.
        """
        result = "/usr/bin/xrdcp -d 0 "
        if options != None:
            result += " %s " % options
        result += " %s " % sourcePFN
        result += " %s " % targetPFN
        result += """
        EXIT_STATUS=$?
        if [[ $EXIT_STATUS != 0 ]]; then
            echo "ERROR: xrdcp exited with $EXIT_STATUS"
        fi
        exit $EXIT_STATUS
        """
        return result

    def removeFile(self, pfnToRemove):
        """
        _removeFile_

        CleanUp pfn provided
        """

        method = self.storageMethod(pfnToRemove)

        if method == 'xrdcp':
            (_, host, path, _) = self.splitPFN(pfnToRemove)
            command = "xrd %s rm %s" % (host, path)
            print("Executing: %s" % command)
            self.executeCommand(command)
        elif method == 'srm':
            return self.srmImpl.removeFile(pfnToRemove)
        else:
            command = "/bin/rm %s" % stripPrefixTOUNIX(pfnToRemove)
            print("Executing: %s" % command)
            self.executeCommand(command)


registerStageOutImpl("stageout-xrdcp-fnal", FNALImpl)



#!/usr/bin/env python
"""
_AnalyticsDataCollector_

Collects request/jobs summary data from local couchdb
and put in local summary couch db.
Then summary data will be replicated to central wmstats couchdb
"""


import threading

from WMCore.Agent.Harness import Harness
from WMComponent.AnalyticsDataCollector.AnalyticsPoller import AnalyticsPoller

class AnalyticsDataCollector(Harness):
    """
    Component class for AnalyticsDataCollector module
    """

    def __init__(self, config):
        """
        __init__

        Initialize the Harness
        """
        Harness.__init__(self, config)
        return

    def preInitialization(self):
        """
        _preInitialization_

        Sets up the workers threads
        """
        pollInterval = self.config.AnalyticsDataCollector.AnalyticsPoller.pollInterval

        myThread = threading.currentThread()

        myThread.workerThreadManager.addWorker(AnalyticsPoller(self.config),
                                               pollInterval)



#!/usr/bin/env python


"""
Checks for finished subscriptions
Upon finding finished subscriptions, notifies WorkQueue and kills them
"""
from __future__ import print_function




import logging
import threading

from WMCore.Agent.Harness import Harness
#from WMCore.WMFactory import WMFactory

from WMComponent.JobTracker.JobTrackerPoller import JobTrackerPoller



class JobTracker(Harness):
    """
    Checks for finished subscriptions
    Upon finding finished subscriptions, notifies WorkQueue and kills them

    """

    def __init__(self, config):
        # call the base class
        Harness.__init__(self, config)
        self.pollTime = 1

        self.config = config

        print("JobTracker.__init__")

    def preInitialization(self):
        """
        Sets up the worker thread

        """
        logging.info("JobTracker.preInitialization")

        # Add event loop to worker manager
        myThread = threading.currentThread()

        pollInterval = self.config.JobTracker.JobTrackerPoller.pollInterval
        logging.info("Setting poll interval to %s seconds" %pollInterval)
        myThread.workerThreadManager.addWorker(JobTrackerPoller(self.config), pollInterval)

        return



#!/usr/bin/env python
"""
_CheckForDelete_

Oracle implementation of DeleteCheck

"""
__all__ = []



from WMCore.WMBS.MySQL.Fileset.CheckForDelete import CheckForDelete as MySQLCheckForDelete

class CheckForDelete(MySQLCheckForDelete):
    """
    Oracle implementation

    """



"""
_InsertComponent_

Oracle implementation of InsertComponent
"""

__all__ = []



from WMCore.Agent.Database.MySQL.InsertComponent import InsertComponent \
     as InsertComponentMySQL

class InsertComponent(InsertComponentMySQL):
    pass



#!/usr/bin/env python
"""
_DeleteFileset_

Oracle implementation of Fileset.Delete

"""
__all__ = []




from WMCore.WMBS.MySQL.Fileset.Delete import Delete as DeleteFilesetMySQL

class Delete(DeleteFilesetMySQL):
    sql = DeleteFilesetMySQL.sql



#!/usr/bin/env python
"""
_UpdateThresholdsBulk_

This module updates thresholds for the given sites for MySQL.
"""




from WMCore.Database.DBFormatter import DBFormatter

class UpdateThresholdsInBulk(DBFormatter):
    """
    _UpdateThresholdsBulk_
    """
    sql = """UPDATE rc_threshold SET max_slots = :maxslots, pending_slots = :pendslots
                WHERE site_id = (SELECT id FROM wmbs_location WHERE site_name = :sitename) AND
                      sub_type_id = (SELECT id FROM wmbs_sub_types WHERE name  = :tasktype)"""


    def execute(self, data, conn = None, transaction = False):

        """
        data is the list of dictionary. each dictionary contains following items.
        [{"sitename": T2_XX_SiteA, "tasktype": Processing, "maxslots": 1000}, ...]
        """
        self.dbi.processData(self.sql, data,
                             conn = conn, transaction = transaction)
        return



#!/usr/bin/env python
"""
_CountFilesByStatus_

MySQL implementation of DBSBuffer.CountFilesByStatus

"""

from __future__ import division
from WMCore.Database.DBFormatter import DBFormatter


class CountFilesByStatus(DBFormatter):
    """
    CountFilesByStatus

    Count the number of files in dbs based on their status

    """
    sql = """SELECT COUNT(*)
             FROM dbsbuffer_file
             WHERE status = :status"""

    def execute(self, status, conn=None, transaction=False):
        binds = {'status': status}

        result = self.dbi.processData(self.sql, binds, conn=conn,
                                      transaction=transaction)
        result = self.format(result)
        return result[0][0]



#!/usr/bin/env python
"""
pullWork poller
"""
import random
import time
import logging

from Utils.Timers import timeFunction
from Utils.wmcoreDTools import resetWatchdogTimer, moduleName
from WMCore.Services.ReqMgr.ReqMgr import ReqMgr
from WMCore.WorkQueue.WMBSHelper import freeSlots
from WMCore.WorkQueue.WorkQueueUtils import cmsSiteNames
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread


class WorkQueueManagerWMBSFileFeeder(BaseWorkerThread):
    """
    Polls for Work
    """

    def __init__(self, queue, config):
        """
        Initialise class members
        """
        BaseWorkerThread.__init__(self)

        self.queue = queue
        self.config = config
        self.reqmgr2Svc = ReqMgr(self.config.General.ReqMgr2ServiceURL)
        # state lists which shouldn't be populated in wmbs. (To prevent creating work before WQE status updated)
        self.abortedAndForceCompleteWorkflowCache = self.reqmgr2Svc.getAbortedAndForceCompleteRequestsFromMemoryCache()

    def setup(self, parameters):
        """
        Called at startup - introduce random delay
             to avoid workers all starting at once
        """
        t = random.randrange(self.idleTime)
        self.logger.info('Sleeping for %d seconds before 1st loop', t)
        time.sleep(t)

    @timeFunction
    def algorithm(self, parameters):
        """
        Get work from local workqueue to be injected into WMBS/DBSBuffer
        """
        self.queue.logger.info("Getting work and feeding WMBS files...")
        try:
            # need to make sure jobs are created
            resources, jobCounts = freeSlots(minusRunning=True, allowedStates=['Normal', 'Draining'],
                                             knownCmsSites=cmsSiteNames())

            for site in resources:
                self.queue.logger.info("I need %d jobs on site %s" % (resources[site], site))

            abortedAndForceCompleteRequests = self.abortedAndForceCompleteWorkflowCache.getData()

            previousWorkList = self.queue.getWork(resources, jobCounts,
                                                  excludeWorkflows=abortedAndForceCompleteRequests)
            self.queue.logger.info("Acquired %s units of work for WMBS file creation", len(previousWorkList))
        except Exception as ex:
            self.queue.logger.error("Error in wmbs inject loop: %s" % str(ex))

        # Reset its own watchdog timer at the end of the run cycle
        logging.info(f"Resetting {moduleName(self)} watchdog timer.")
        if resetWatchdogTimer(self):
            logging.warning(f"Failed to reset {moduleName(self)} watchdog timer. The component might be restarted soon.")



#!/usr/bin/env python


"""
Checks for finished subscriptions
Upon finding finished subscriptions, notifies WorkQueue and kills them
"""
from __future__ import print_function




import logging
import threading

from WMCore.Agent.Harness import Harness
from WMCore.WMFactory import WMFactory

from WMComponent.TaskArchiver.TaskArchiverPoller import TaskArchiverPoller
from WMComponent.TaskArchiver.CleanCouchPoller import CleanCouchPoller

class TaskArchiver(Harness):

    def __init__(self, config):
        # call the base class
        Harness.__init__(self, config)
        self.pollTime = 1

        self.config = config

        print("TaskArchiver.__init__")

    def preInitialization(self):
        print("TaskArchiver.preInitialization")

        # Add event loop to worker manager
        myThread = threading.currentThread()

        pollInterval = self.config.TaskArchiver.TaskArchiverPoller.pollInterval
        logging.info("Setting task archiver poll interval to %s seconds" % pollInterval)
        myThread.workerThreadManager.addWorker(TaskArchiverPoller(self.config), pollInterval)

        couchInterval = self.config.TaskArchiver.CleanCouchPoller.pollInterval
        logging.info("Setting poll interval for cleanup old couch doc to %s seconds" % couchInterval)
        myThread.workerThreadManager.addWorker(CleanCouchPoller(self.config),
                                               couchInterval)

        return



#!/usr/bin/env python
"""
_AddToFileset_

MySQL implementation of Files.AddToFileset
"""

import time

from WMCore.Database.DBFormatter import DBFormatter

class AddToFileset(DBFormatter):
    sql = """INSERT IGNORE INTO wmbs_fileset_files (fileid, fileset, insert_time)
               SELECT wmbs_file_details.id, :fileset, :insert_time
               FROM wmbs_file_details
               WHERE wmbs_file_details.lfn = :lfn
               """

    sqlAvail = """INSERT IGNORE INTO wmbs_sub_files_available (subscription, fileid)
                    SELECT wmbs_subscription.id AS subscription,
                           wmbs_file_details.id AS fileid FROM wmbs_subscription
                      INNER JOIN wmbs_file_details ON
                        wmbs_file_details.lfn = :lfn
                    WHERE wmbs_subscription.fileset = :fileset
                    """

    def execute(self, file = None, fileset = None, conn = None,
                transaction = False):
        binds = []
        availBinds = []
        timestamp = int(time.time())
        for fileLFN in file:
            binds.append({"lfn": fileLFN, "fileset": fileset,
                          "insert_time": timestamp})
            availBinds.append({"lfn": fileLFN, "fileset": fileset})

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        self.dbi.processData(self.sqlAvail, availBinds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python

"""
Oracle implementation of AddCKType
"""





from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.AddCKType import AddCKType as MySQLAddCKType

class AddCKType(MySQLAddCKType):
    sql = """INSERT INTO dbsbuffer_checksum_type (type)
             VALUES (:cktype)"""



#!/usr/bin/env python

"""
_ErrorHandler_

The error handler pools for error conditions
(CreateFailed, SubmitFailed, and JobFailed)
By looking at wmbs_job table's status filed.
All the jobs are handled respectively.

the different failure handlers are configurable in the config file and
relate to the three stages of a job: create, submit, run

The component runs in Poll mode, basically submits itself
'Poll' message at the end of each cycle, so that it keeps polling
We can introduce some delay in polling, if have to.
"""





import logging
import threading

# harness class that encapsulates the basic component logic.
from WMCore.Agent.Harness import Harness


from WMComponent.ErrorHandler.ErrorHandlerPoller import ErrorHandlerPoller

class ErrorHandler(Harness):
    """
    _ErrorHandler_

    The error handler pools for error conditions (CreateFailed, SubmitFailed, and JobFailed)
    By looking at wmbs_job table's status filed.
    All the errors are handled respectively by handlers related to
    the three stages of a job: create, submit, run
    """

    def __init__(self, config):
        # call the base class
        Harness.__init__(self, config)
        self.config = config

    def preInitialization(self):
        """
        Initializes plugins for different messages
        """

        # Add event loop to worker manager
        myThread = threading.currentThread()
        pollInterval = self.config.ErrorHandler.ErrorHandlerPoller.pollInterval
        logging.info("Setting poll interval to %s seconds" % pollInterval)
        myThread.workerThreadManager.addWorker(ErrorHandlerPoller(self.config), pollInterval)



#!/usr/bin/env python
"""
_Save_

MySQL implementation of Jobs.Save
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Save(DBFormatter):
    sql = """UPDATE wmbs_job SET jobgroup = :jobgroup, name = :name,
               couch_record = :couch_record, outcome = :outcome, cache_dir = :cache_dir,
               location =
                 (SELECT id FROM wmbs_location WHERE site_name = :location),
               fwjr_path = :fwjr, retry_count = :retry_count
             WHERE id = :jobid"""

    def execute(self, jobid, jobgroup, name, couch_record, location, outcome, cache_dir, fwjr,
                retry_count = 0, conn = None, transaction = False):
        if outcome == 'success':
            boolOutcome = 1
        else:
            boolOutcome = 0

        binds = {"jobid": jobid, "jobgroup": jobgroup, "name": name,
                 "couch_record": couch_record, "location": location,
                 "outcome": boolOutcome, "cache_dir": cache_dir, "fwjr": fwjr,
                 'retry_count': retry_count}

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_Template_

Template class for all Step Template implementations to inherit and implement
the API

"""
from builtins import object
import os

from WMCore.WMSpec.WMStep import WMStepHelper
from WMCore.WMSpec.ConfigSectionTree import nodeName

class CoreHelper(WMStepHelper):
    """
    _CoreHelper_

    Helper API for core settings

    """
    def stepName(self):
        """
        _stepName_

        Get the name of the step

        """
        return nodeName(self.data)


    def addEnvironmentVariable(self, varname, setting):
        """
        _addEnvironmentVariable_

        add a key = value style setting to the environment for this
        step

        """
        setattr(self.data.environment.variables, varname, setting)
        return

    def addEnvironmentPath(self, pathname, setting):
        """
        _addEnvironmentPath_

        add a key = value1:value2:value3 environment setting to this step

        """
        if getattr(self.data.environment.paths, pathname, None) == None:
            setattr(self.data.environment.paths, pathname, [])
        pathentry = getattr(self.data.environment.paths, pathname)
        pathentry.append(setting)
        return

    def environment(self):
        """
        _environment_

        Get the environment settings for this step

        """
        return self.data.environment

    def setOverrideCatalog(self, overrideCatalog):
        """
        _setOverrideCatalog_
        set the override catalog needed at least at CERN to use production castor pools
        """
        if overrideCatalog is not None:
            self.data.application.overrideCatalog = overrideCatalog

    def getOverrideCatalog(self):
        """
        _getOverrideCatalog_
        return the TFC specified in overrideCatalog.
        """
        return getattr(self.data.application, "overrideCatalog", None)

    def addDirectory(self, dirName):
        """
        _addDirectory_

        Add a subdirectory structure to the template that will be built by
        the builder

        """
        split = dirName.split("/")
        split = [ x for x in split if x.strip() != "" ]

        dirs = getattr(self.data.build.directories, self.stepName())
        for subdir in split:
            exists = getattr(dirs, subdir, None)
            if exists == None:
                dirs.section_(subdir)
            dirs = getattr(dirs, subdir)
        return dirs



    def addFile(self, fileName, newLocation = None):
        """
        _addFile_

        Add a file to the job at build time. This file must be
        a local filesystem file available at fileName.

        An optional location within the step can be specified which
        may include a path structure that gets translated into calls
        to addDirectory

        """
        dirs = getattr(self.data.build.directories, self.stepName())
        if newLocation != None:
            filename = os.path.basename(newLocation)
            dirname = os.path.dirname(newLocation)
            dirs = self.addDirectory(dirname)
            setattr(dirs, filename, { "Source" : fileName, "Target" : filename})
        else:
            filename = os.path.basename(fileName)
            setattr(dirs, filename, {"Target" : filename, "Source" : fileName })
        return


    def directoryStructure(self):
        """
        _directoryStructure_

        Util to retrieve the directory structure

        """
        return self.data.build.directories

class Template(object):
    """
    _Template_

    Base interface definition for any WMStep Template

    """

    def __init__(self):
        pass

    def __call__(self, step):
        """
        _operator(step)_

        Install the template on the step instance provided

        """
        self.coreInstall(step)
        self.install(step)


    def coreInstall(self, step):
        """
        _coreInstall_

        Install attributes common to all steps

        """
        # Environment settings to pass to the step
        step.section_("environment")
        step.environment.section_("variables")
        step.environment.section_("paths")

        # Directory structure and files to be included in the job
        # beyond those that would be added by a Step Specific builder
        # Step Specific subclasses can simply append to these to get files
        # and dirs into the job
        step.section_("build")
        step.build.section_("directories")
        step.build.directories.section_(nodeName(step))






    def install(self, step):
        """
        _install_

        Override this method to install the required attributes
        in the step Instance provided

        """
        msg = "WMSpec.Steps.Template.install method not overridden in "
        msg += "implementation: %s\n" % self.__class__.__name__
        raise NotImplementedError(msg)


    def helper(self, step):
        """
        _helper_

        Wrap the step instance in a helper class tailored to this particular
        step type

        """
        msg = "WMSpec.Steps.Template.helper method not overridden in "
        msg += "implementation: %s\n" % self.__class__.__name__
        raise NotImplementedError(msg)



#!/usr/bin/env python
"""
_XRDCPImpl_

Implementation of StageOutImpl interface for xrdcp

Generic, will/should work with any site.

"""
from __future__ import print_function

import argparse
import logging
import os

from WMCore.Storage.Registry import registerStageOutImpl
from WMCore.Storage.StageOutImpl import StageOutImpl
from WMCore.Storage.StageOutError import StageOutError

class XRDCPImpl(StageOutImpl):
    """
    _XRDCPImpl_

    """

    def __init__(self, stagein=False):
        StageOutImpl.__init__(self, stagein)
        self.numRetries = 5
        self.retryPause = 300
        self.xrdfsCmd = "xrdfs"
        self.setAuthX509 = "env X509_USER_PROXY=$X509_USER_PROXY "
        # if BEARER_TOKEN_FILE is not set, use /dev/null as a fallback to suppress the error
        self.setAuthToken = "env BEARER_TOKEN_FILE=$BEARER_TOKEN_FILE BEARER_TOKEN=$(cat ${BEARER_TOKEN_FILE:-/dev/null}) "
        self.unsetX509 = "X509_USER_PROXY= "
        self.unsetToken = "BEARER_TOKEN_FILE= BEARER_TOKEN= "
        self.debuggingTemplate = "#!/bin/bash\n"
        self.debuggingTemplate += """
        echo
        echo
        echo "-----------------------------------------------------------"
        echo "==========================================================="
        echo
        echo "Debugging information on failing xrdcp/xrdfs command"
        echo
        echo "Current date and time: $(date +"%Y-%m-%d %H:%M:%S")"
        echo "XRootD command which failed: {copy_command}"
        echo "Hostname:   $(hostname -f)"
        echo "OS:  $(uname -r -s)"
        echo
        echo "XRD environment variables (if any):"
        env | grep ^XRD_
        echo
        echo "PYTHON environment variables:"
        env | grep ^PYTHON
        echo
        echo "LD_* environment variables:"
        env | grep ^LD_
        echo
        echo "xrdcp location: $(which xrdcp)"
        echo "xrdfs location: $(which xrdfs)"
        echo "Source PFN: {source}"
        echo "Target PFN: {destination}"
        echo
        echo
        echo "Information for credentials in the environment"
        echo "Bearer token content: $BEARER_TOKEN"
        echo "Bearer token file: $BEARER_TOKEN_FILE"
        echo
        echo "VOMS proxy info:"
        voms-proxy-info -all
        echo "==========================================================="
        echo "-----------------------------------------------------------"
        echo
        """
    def createSourceName(self, protocol, pfn):
        """
        _createSourceName_

        """
        return pfn

    def createOutputDirectory(self, targetPFN):
        """
        _createOutputDirectory_

        not needed since xrdcp does it automatically
        """
        return

    def checkXRDUtilsExist(self):
        """
        Verifies whether xrdcp and xrdfs utils exist in the job path.

        :return: True if both exist, otherwise False
        """
        foundXrdcp, foundXrdfs = False, False
        for path in os.environ["PATH"].split(os.pathsep):
            if os.access(os.path.join(path, "xrdcp"), os.X_OK):
                foundXrdcp = True
            if os.access(os.path.join(path, "xrdfs"), os.X_OK):
                foundXrdfs = True

        return foundXrdcp & foundXrdfs

    def _getAuthEnv(self, authMethod=None, forceMethod=False):
        """
        Get environment variables for stageout command based on the selected authentication method.
        :authMethod: str, the authentication method to be used ("X509", "TOKEN", or None)
        :forceMethod: bool, cleans non-chosen auth methods from environment.
        :return: str
        """
        authEnv = ""
        if authMethod is None:
            return authEnv

        if authMethod.upper() == 'TOKEN' and forceMethod is False:
            if not self.isBearerTokenFileSet():
                msg = "Stage out requested with tokens, but environment variable is not defined."
                msg += " Forcing it to use X509 authentication method instead."
                logging.info(msg)
                authMethod = 'X509'
        if authMethod.upper() == 'TOKEN':
            authEnv = self.setAuthToken
            if forceMethod:
                authEnv += self.unsetX509
        elif authMethod.upper() == 'X509':
            authEnv = self.setAuthX509
            if forceMethod:
                authEnv += self.unsetToken
        else:
            logging.warning("Warning! Running without either a X509 certificate or a token specified!")

        return authEnv

    def getFormattedCopyCommand(self, sourcePFN, targetPFN, options=None, checksums=None, authMethod=None, forceMethod=False, dryRun=False):
        """
        Construct xrdcp/xrdfs stageout command.
        If adler32 checksum is provided, use it for the transfer
        xrdcp options used:
          --force : re-creates a file if it's already present
          --nopbar : does not display the progress bar

        :sourcePFN: str, the source PFN.
        :targetPFN: str, the target PFN.
        :options: str, additional options for gfal-copy.
        :checksums: dict, checksum values.
        :authMethod: str, preferred authentication method.
        :forceMethod: bool, whether to force the preferred authentication method.
        :dryRun: bool, dry run mode (to enable debug mode)
        """
        if not options:
            options = ''

        parser = argparse.ArgumentParser()
        parser.add_argument('--wma-cerncastor', action='store_true')
        parser.add_argument('--wma-disablewriterecovery', action='store_true')
        parser.add_argument('--wma-preload')
        args, unknown = parser.parse_known_args(options.split())

        # strip out WMAgent specific options
        unknown = [option for option in unknown if not option.startswith('--wma-')]
        copyCommandOptions = ' '.join(unknown)

        copyCommand = ""

        if self.stageIn:
            remotePFN, localPFN = sourcePFN, targetPFN
        else:
            remotePFN, localPFN = targetPFN, sourcePFN
            copyCommand += "LOCAL_SIZE=`stat -c%%s \"%s\"`\n" % localPFN
            copyCommand += "echo \"Local File Size is: $LOCAL_SIZE\"\n"
            if args.wma_cerncastor:
                targetPFN += "?svcClass=t0cms"

        useChecksum = (checksums is not None and 'adler32' in checksums and not self.stageIn)

        # If not, check whether OSG cvmfs repo is available.
        # This likely only works on RHEL7 x86-64 (and compatible) OS, but this
        # represents most of our resources and it's a fallback mechanism anyways
        if not self.checkXRDUtilsExist():
            logging.warning("Failed to find XRootD in the path. Trying fallback to OSG CVMFS...")
            initFile = "/cvmfs/oasis.opensciencegrid.org/osg-software/osg-wn-client/current/el7-x86_64/setup.sh"
            if os.path.isfile(initFile):
                copyCommand += "source %s\n" % initFile

        if args.wma_disablewriterecovery:
            copyCommand += "env XRD_WRITERECOVERY=0 "

        if args.wma_preload:
            xrdcpCmd = "%s xrdcp" % args.wma_preload
            self.xrdfsCmd = "%s xrdfs" % args.wma_preload
        else:
            xrdcpCmd = "xrdcp"
            self.xrdfsCmd = "xrdfs"

        # Check for auth-related environment to prepend
        authEnv = self._getAuthEnv(authMethod, forceMethod)
        if authEnv:
            copyCommand += authEnv
            self.xrdfsCmd = "%s %s" % (authEnv, self.xrdfsCmd)
        copyCommand += "%s --force --nopbar " % xrdcpCmd
        
        if copyCommandOptions:
            copyCommand += "%s " % copyCommandOptions

        if useChecksum:
            checksums['adler32'] = "%08x" % int(checksums['adler32'], 16)
            copyCommand += "--cksum adler32:%s " % checksums['adler32']

        copyCommand += " \"%s\" " % sourcePFN
        copyCommand += " \"%s\" \n" % targetPFN
        copyCommand += "RC=$? \n"
        copyCommand += "echo \"xrdcp exit code: $RC\" \n"
        if self.stageIn:
            copyCommand += "LOCAL_SIZE=`stat -c%%s \"%s\"`\n" % localPFN
            copyCommand += "echo \"Local File Size is: $LOCAL_SIZE\"\n"

        (_, host, path, _) = self.splitPFN(remotePFN)

        if self.stageIn:
            removeCommand = ""
        else:
            removeCommand = "%s '%s' rm %s ;" % (self.xrdfsCmd, host, path)

        copyCommand += "REMOTE_SIZE=`%s '%s' stat '%s' | grep Size | sed -r 's/.*Size:[ ]*([0-9]+).*/\\1/'`\n" % (self.xrdfsCmd, host, path)
        copyCommand += "echo \"Remote File Size is: $REMOTE_SIZE\"\n"

        if useChecksum:

            copyCommand += "echo \"Local File Checksum is: %s\"\n" % checksums['adler32']
            copyCommand += "REMOTE_XS=`%s '%s' query checksum '%s' | grep -i adler32 | sed -r 's/.*[adler|ADLER]32[ ]*([0-9a-fA-F]{8}).*/\\1/'`\n" % (self.xrdfsCmd, host, path)
            copyCommand += "echo \"Remote File Checksum is: $REMOTE_XS\"\n"

            copyCommand += "if [ $RC == 0 ] && [ $REMOTE_SIZE ] && [ $REMOTE_XS ] && [ $LOCAL_SIZE == $REMOTE_SIZE ] && [ '%s' == $REMOTE_XS ]; then exit 0; " % checksums['adler32']
            copyCommand += "else echo \"ERROR: XRootD file transfer return code is $RC. Size or Checksum Mismatch between local and SE\"; %s exit 60311 ; fi" % removeCommand

        else:

            copyCommand += "if [ $RC == 0 ] && [ $REMOTE_SIZE ] && [ $LOCAL_SIZE == $REMOTE_SIZE ]; then exit 0; "
            copyCommand += "else echo \"ERROR: XRootD file transfer return code is $RC. Size or Checksum Mismatch between local and SE\"; %s exit 60311 ; fi" % removeCommand

        if dryRun:
            return self.debuggingTemplate.format(copy_command=copyCommand, source=sourcePFN, destination=targetPFN)

        return copyCommand

    def createStageOutCommand(self, sourcePFN, targetPFN, options=None, checksums=None, authMethod=None, forceMethod=False):
        """
        _createStageOutCommand_

        Build the actual xrdcp stageout command

        :sourcePFN: str, the source PFN.
        :targetPFN: str, the target PFN.
        :options: str, additional options for gfal-copy.
        :checksums: dict, checksum values.
        :authMethod: str, preferred authentication method.
        :forceMethod: bool, whether to force the preferred authentication method.
        :dryRun: bool, dry run mode (to enable debug mode).

        """
        # Construct xrdcp stageout command and return it
        return self.getFormattedCopyCommand(sourcePFN, targetPFN, options, checksums, authMethod, forceMethod)
    
    def createDebuggingCommand(self, sourcePFN, targetPFN, options=None, checksums=None, authMethod=None, forceMethod=False):
        """
        Debug a failed xrdcp command for stageOut, without re-running it,
        providing information on the environment and the certifications

        :sourcePFN: str, PFN of the source file
        :targetPFN: str, destination PFN
        :options: str, additional options for copy command
        :checksums: dict, collect checksums according to the algorithms saved as keys
        :authMethod: str, the authentication method to be preferentially used ("X509", "TOKEN", or None)
        :forceMethod: bool, whether to force the use of the preferred authentication method, disabling the other 
        """
        # Construct xrdcp/xrdfs commands, but returning the debugging information instead of running it
        return self.getFormattedCopyCommand(sourcePFN, targetPFN, options, checksums, authMethod, forceMethod, dryRun=True)

    def removeFile(self, pfnToRemove):
        """
        Remove a file
        :pfnToRemove: str, PFN of the source file
        """
        (_, host, path, _) = self.splitPFN(pfnToRemove)

        # Try TOKEN authentication first, follow with safe-logic auth methods otherwise
        authEnv = self._getAuthEnv(authMethod="TOKEN")
        command = "%s %s %s rm %s" % (authEnv, self.xrdfsCmd, host, path)

        try:
            self.executeCommand(command)
            logging.info("Removing file with TOKEN authentication succeeded (X509 still enabled).")
        except StageOutError as ex:
            logging.error("Removing file with TOKEN authentication failed. Command: %s", command)
            logging.error("Error: %s", str(ex))
            logging.info("Attempting to remove file with authentication safe-logic")
            if os.getenv("X509_USER_PROXY"):
                logging.info("Retrying with X509_USER_PROXY with BEARER_TOKEN unset...")
                authEnv = self._getAuthEnv(authMethod="X509", forceMethod=True)
                command = "%s %s %s rm %s" % (authEnv, self.xrdfsCmd, host, path)
                try:
                    self.executeCommand(command)
                    logging.info("removefile succeeded with X509 with BEARER_TOKEN unset.")
                    return
                except StageOutError as fallbackEx:
                    logging.error("Fallback with X509_USER_PROXY failed:\n%s", str(fallbackEx))

            if os.getenv("BEARER_TOKEN") or os.getenv("BEARER_TOKEN_FILE"):
                logging.info("Retrying with BEARER_TOKEN with X509_USER_PROXY unset...")
                authEnv = self._getAuthEnv(authMethod="TOKEN", forceMethod=True)
                command = "%s %s %s rm %s" % (authEnv, self.xrdfsCmd, host, path)
                try:
                    self.executeCommand(command)
                    logging.info("Removing file succeeded with TOKEN with X509_USER_PROXY unset.")
                    return
                except StageOutError as fallbackEx:
                    logging.error("Fallback with BEARER_TOKEN failed:\n%s", str(fallbackEx))
              
        return

registerStageOutImpl("xrdcp", XRDCPImpl)



#!/usr/bin/env python
"""
_CMSCouch_

A simple API to CouchDB that sends HTTP requests to the REST interface.

http://wiki.apache.org/couchdb/API_Cheatsheet

NOT A THREAD SAFE CLASS.
"""
from __future__ import print_function, division
from builtins import str as newstr, bytes as newbytes, object
from Utils.Utilities import decodeBytesToUnicode, encodeUnicodeToBytes, decodeBytesToUnicodeConditional
from Utils.PythonVersion import PY3

from future import standard_library
standard_library.install_aliases()
from future.utils import viewitems
import urllib.request, urllib.parse, urllib.error

import base64
import hashlib
import json
import logging
import re
import time
import sys
from pprint import pformat
from datetime import datetime
from http.client import HTTPException

from Utils.IteratorTools import grouper, nestedDictUpdate
from WMCore.Lexicon import sanitizeURL
from WMCore.Services.Requests import JSONRequests


def check_name(dbname):
    match = re.match("^[a-z0-9_$()+-/]+$", urllib.parse.unquote_plus(dbname))
    if not match:
        msg = '%s is not a valid database name'
        raise ValueError(msg % urllib.parse.unquote_plus(dbname))


def check_server_url(srvurl):
    good_name = srvurl.startswith('http://') or srvurl.startswith('https://')
    if not good_name:
        raise ValueError('You must include http(s):// in your servers address')


PY3_STR_DECODER = lambda x: decodeBytesToUnicodeConditional(x, condition=PY3)


class Document(dict):
    """
    Document class is the instantiation of one document in the CouchDB
    """

    def __init__(self, id=None, inputDict=None):
        """
        Initialise our Document object - a dictionary which has an id field
        inputDict - input dictionary to initialise this instance
        """
        inputDict = inputDict or {}
        dict.__init__(self)
        self.update(inputDict)
        if id:
            self.setdefault("_id", id)

    def delete(self):
        """
        Mark the document as deleted
        """
        # https://issues.apache.org/jira/browse/COUCHDB-1141
        deletedDict = {'_id': self['_id'], '_rev': self['_rev'], '_deleted': True}
        self.update(deletedDict)
        for key in list(self.keys()):
            if key not in deletedDict:
                del self[key]

    def __to_json__(self, thunker):
        """
        __to_json__

        This is here to prevent the serializer from attempting to serialize
        this object and adding a bunch of keys that couch won't understand.
        """
        jsonDict = {}
        for key in self.keys():
            jsonDict[key] = self[key]

        return jsonDict


class CouchDBRequests(JSONRequests):
    """
    CouchDB has two non-standard HTTP calls, implement them here for
    completeness, and talks to the CouchDB port
    """

    def __init__(self, url='http://localhost:5984', usePYCurl=True, ckey=None, cert=None, capath=None):
        """
        Initialise requests
        """
        JSONRequests.__init__(self, url,
                              {"cachepath": None, "pycurl": usePYCurl, "key": ckey, "cert": cert, "capath": capath})
        self.accept_type = "application/json"
        self["timeout"] = 600

    def move(self, uri=None, data=None):
        """
        MOVE some data
        """
        return self.makeRequest(uri, data, 'MOVE')

    def copy(self, uri=None, data=None):
        """
        COPY some data
        """
        return self.makeRequest(uri, data, 'COPY')

    def makeRequest(self, uri=None, data=None, type='GET', incoming_headers=None,
                    encode=True, decode=True, contentType=None, cache=False):
        """
        Make the request, handle any failed status, return just the data (for
        compatibility). By default do not cache the response.

        TODO: set caching in the calling methods.
        """
        incoming_headers = incoming_headers or {}
        incoming_headers.update(self.additionalHeaders)
        try:
            if not cache:
                incoming_headers.update({'Cache-Control': 'no-cache'})
            result, status, reason, cached = JSONRequests.makeRequest(
                    self, uri, data, type, incoming_headers,
                    encode, decode, contentType)
        except HTTPException as e:
            self.checkForCouchError(getattr(e, "status", None),
                                    getattr(e, "reason", None),
                                    data,
                                    getattr(e, "result", None))

        return result

    def checkForCouchError(self, status, reason, data=None, result=None):
        """
        _checkForCouchError_

        Check the HTTP status and raise an appropriate exception.
        """
        if status == 400:
            raise CouchBadRequestError(reason, data, result, status)
        elif status == 401:
            raise CouchUnauthorisedError(reason, data, result, status)
        elif status == 403:
            raise CouchForbidden(reason, data, result, status)
        elif status == 404:
            raise CouchNotFoundError(reason, data, result, status)
        elif status == 405:
            raise CouchNotAllowedError(reason, data, result, status)
        elif status == 406:
            raise CouchNotAcceptableError(reason, data, result, status)
        elif status == 409:
            raise CouchConflictError(reason, data, result, status)
        elif status == 410:
            raise CouchFeatureGone(reason, data, result, status)
        elif status == 412:
            raise CouchPreconditionFailedError(reason, data, result, status)
        elif status == 413:
            raise CouchRequestTooLargeError(reason, data, result, status)
        elif status == 416:
            raise CouchRequestedRangeNotSatisfiableError(reason, data, result, status)
        elif status == 417:
            raise CouchExpectationFailedError(reason, data, result, status)
        elif status == 500:
            raise CouchInternalServerError(reason, data, result, status)
        elif status in [502, 503, 504]:
            # There are HTTP errors that CouchDB doesn't raise but can appear
            # in our environment, e.g. behind a proxy. Reraise the HTTPException
            raise CouchError(reason, data, result, status)
        else:
            # We have a new error status, log it
            raise CouchError(reason, data, result, status)



class Database(CouchDBRequests):
    """
    Object representing a connection to a CouchDB Database instance.
    TODO: implement COPY and MOVE calls.
    TODO: remove leading whitespace when committing a view
    """

    def __init__(self, dbname='database', url='http://localhost:5984', size=1000, ckey=None, cert=None):
        """
        A set of queries against a CouchDB database
        """
        check_name(dbname)

        self.name = urllib.parse.quote_plus(dbname)

        CouchDBRequests.__init__(self, url=url, ckey=ckey, cert=cert)
        self._reset_queue()

        self._queue_size = size
        self.threads = []
        self.last_seq = 0

    def _reset_queue(self):
        """
        Set the queue to an empty list, e.g. after a commit
        """
        self._queue = []

    def timestamp(self, data, label=''):
        """
        Time stamp each doc in a list
        """
        if label is True:
            label = 'timestamp'

        if isinstance(data, type({})):
            data[label] = int(time.time())
        else:
            for doc in data:
                if label not in doc:
                    doc[label] = int(time.time())
        return data

    def getQueueSize(self):
        """
        Return the current size of the queue, i.e., how
        many documents are already queued up
        """
        return len(self._queue)

    def queue(self, doc, timestamp=False, viewlist=None, callback=None):
        """
        Queue up a doc for bulk insert. If timestamp = True add a timestamp
        field if one doesn't exist. Use this over commit(timestamp=True) if you
        want to timestamp when a document was added to the queue instead of when
        it was committed
        If a callback is specified then pass it to the commit function if a
        commit is triggered
        """
        viewlist = viewlist or []
        if timestamp:
            self.timestamp(doc, timestamp)
        # TODO: Thread this off so that it's non blocking...
        if self.getQueueSize() >= self._queue_size:
            logging.warning('queue larger than %s records, committing', self._queue_size)
            self.commit(viewlist=viewlist, callback=callback)
        self._queue.append(doc)

    def queueDelete(self, doc):
        """
        Queue up a document for deletion
        """
        assert isinstance(doc, type({})), "document not a dictionary"
        # https://issues.apache.org/jira/browse/COUCHDB-1141
        doc = {'_id': doc['_id'], '_rev': doc['_rev'], '_deleted': True}
        self.queue(doc)

    def commitOne(self, doc, timestamp=False, viewlist=None):
        """
        Helper function for when you know you only want to insert one doc
        additionally keeps from having to rewrite ConfigCache to handle the
        new commit function's semantics
        """
        viewlist = viewlist or []
        uri = '/%s/_bulk_docs/' % self.name
        if timestamp:
            self.timestamp(doc, timestamp)

        data = {'docs': [doc]}
        retval = self.post(uri, data)
        for v in viewlist:
            design, view = v.split('/')
            self.loadView(design, view, {'limit': 0})
        return retval

    def commit(self, doc=None, returndocs=False, timestamp=False,
               viewlist=None, callback=None, **data):
        """
        Add doc and/or the contents of self._queue to the database.
        If timestamp is true timestamp all documents with a unix style
        timestamp - this will be the timestamp of when the commit was called, it
        will not override an existing timestamp field.  If timestamp is a string
        that string will be used as the label for the timestamp.

        The callback function will be called with the documents that trigger a
        conflict when doing the bulk post of the documents in the queue,
        callback functions must accept the database object, the data posted and a row in the
        result from the bulk commit. The callback updates the retval with
        its internal retval

        key, value pairs can be used to pass extra parameters to the bulk doc api
        See https://docs.couchdb.org/en/latest/api/database/bulk-api.html#db-bulk-docs

        TODO: restore support for returndocs and viewlist

        Returns a list of good documents
            throws an exception otherwise
        """
        viewlist = viewlist or []
        if doc:
            self.queue(doc, timestamp, viewlist)

        if not self._queue:
            return

        if timestamp:
            self.timestamp(self._queue, timestamp)
        # TODO: commit in thread to avoid blocking others
        uri = '/%s/_bulk_docs/' % self.name

        data['docs'] = list(self._queue)
        retval = self.post(uri, data)
        self._reset_queue()
        for v in viewlist:
            design, view = v.split('/')
            self.loadView(design, view, {'limit': 0})
        if callback:
            for idx, result in enumerate(retval):
                if result.get('error', None) == 'conflict':
                    retval[idx] = callback(self, data, result)

        return retval

    def document(self, id, rev=None):
        """
        Load a document identified by id. You can specify a rev to see an older revision
        of the document. This **should only** be used when resolving conflicts, relying
        on CouchDB revisions for document history is not safe, as any compaction will
        remove the older revisions.
        """
        uri = '/%s/%s' % (self.name, urllib.parse.quote_plus(id))
        if rev:
            uri += '?' + urllib.parse.urlencode({'rev': rev})
        return Document(id=id, inputDict=self.get(uri))

    def updateDocument(self, doc_id, design, update_func, fields=None, useBody=False):
        """
        Call the update function update_func defined in the design document
        design for the document doc_id with a query string built from fields.

        http://wiki.apache.org/couchdb/Document_Update_Handlers
        """
        fields = fields or {}
        # Clean up /'s in the name etc.
        doc_id = urllib.parse.quote_plus(doc_id)

        if not useBody:
            updateUri = '/%s/_design/%s/_update/%s/%s?%s' % \
                        (self.name, design, update_func, doc_id, urllib.parse.urlencode(fields))

            return self.put(uri=updateUri, decode=PY3_STR_DECODER)
        else:
            updateUri = '/%s/_design/%s/_update/%s/%s' % \
                        (self.name, design, update_func, doc_id)
            return self.put(uri=updateUri, data=fields, decode=PY3_STR_DECODER)

    def updateBulkDocuments(self, doc_ids, paramsToUpdate, updateLimits=1000):

        uri = '/%s/_bulk_docs/' % self.name
        conflictDocIDs = []
        for ids in grouper(doc_ids, updateLimits):
            # get original documens
            docs = self.allDocs(options={"include_docs": True}, keys=ids)['rows']
            data = {}
            data['docs'] = []
            for j in docs:
                doc = {}
                doc.update(j['doc'])
                nestedDictUpdate(doc, paramsToUpdate)
                data['docs'].append(doc)

            if data['docs']:
                retval = self.post(uri, data)
                for result in retval:
                    if result.get('error', None) == 'conflict':
                        conflictDocIDs.append(result['id'])

        return conflictDocIDs

    def updateBulkDocumentsWithConflictHandle(self, doc_ids, updateParams, updateLimits=1000, maxConflictLimit=10):
        """
        param: doc_ids: list couch doc ids for updates, shouldn't contain any duplicate or empty string
        param: updateParams: dictionary of parameters to be updated.
        param: updateLimits: number of documents in one commit
        pram maxConflictLimit: number of conflicts fix tries before we give up to fix it to prevent infinite calls
        """
        conflictDocIDs = self.updateBulkDocuments(doc_ids, updateParams, updateLimits)
        if conflictDocIDs:
            # wait a second before trying again for the confict documents
            if maxConflictLimit == 0:
                return conflictDocIDs
            time.sleep(1)
            self.updateBulkDocumentsWithConflictHandle(conflictDocIDs, updateParams,
                                                       maxConflictLimit=maxConflictLimit - 1)
        return []

    def putDocument(self, doc_id, fields):
        """
        Call the update function update_func defined in the design document
        design for the document doc_id with a query string built from fields.

        http://wiki.apache.org/couchdb/Document_Update_Handlers
        """
        # Clean up /'s in the name etc.
        doc_id = urllib.parse.quote_plus(doc_id)

        updateUri = '/%s/%s' % (self.name, doc_id)
        return self.put(uri=updateUri, data=fields, decode=PY3_STR_DECODER)

    def documentExists(self, id, rev=None):
        """
        Check if a document exists by ID. If specified check that the revision rev exists.
        """
        uri = "/%s/%s" % (self.name, urllib.parse.quote_plus(id))
        if rev:
            uri += '?' + urllib.parse.urlencode({'rev': rev})
        try:
            self.makeRequest(uri, {}, 'HEAD')
            return True
        except CouchNotFoundError:
            return False

    def delete_doc(self, id, rev=None):
        """
        Immediately delete a document identified by id and rev.
        If revision is not provided, we need to first fetch this
        document to read the current revision number.

        :param id: string with the document name
        :param rev: string with the revision number
        :return: an empty dictionary if it fails to fetch the document,
            or a dictionary with the deletion outcome, e.g.:
            {'ok': True, 'id': 'doc_name', 'rev': '3-f68156d'}
        """
        uri = '/%s/%s' % (self.name, urllib.parse.quote_plus(id))
        if not rev:
            # then we need to fetch the latest revision number
            doc = self.getDoc(id)
            if "_rev" not in doc:
                logging.warning("Failed to retrieve doc id: %s for deletion.", id)
                return doc
            rev = doc["_rev"]
        uri += '?' + urllib.parse.urlencode({'rev': rev})
        return self.delete(uri)

    def compact(self, views=None, blocking=False, blocking_poll=5, callback=False):
        """
        Compact the database: http://wiki.apache.org/couchdb/Compaction

        If given, views should be a list of design document name (minus the
        _design/ - e.g. myviews not _design/myviews). For each view in the list
        view compaction will be triggered. Also, if the views list is provided
        _view_cleanup is called to remove old view output.

        If True blocking will cause this call to wait until the compaction is
        completed, polling for status with frequency blocking_poll and calling
        the function specified by callback on each iteration.

        The callback function can be used for logging and could also be used to
        timeout the compaction based on status (e.g. don't time out if compaction
        is less than X% complete. The callback function takes the Database (self)
        as an argument. If the callback function raises an exception the block is
        removed and the compact call returns.
        """
        views = views or []
        response = self.post('/%s/_compact' % self.name)
        if views:
            for view in views:
                response[view] = self.post('/%s/_compact/%s' % (self.name, view))
                response['view_cleanup'] = self.post('/%s/_view_cleanup' % (self.name))

        if blocking:
            while self.info()['compact_running']:
                if callback:
                    try:
                        callback(self)
                    except Exception:
                        return response
                time.sleep(blocking_poll)
        return response

    def changes(self, since=-1):
        """
        Get the changes since sequence number. Store the last sequence value to
        self.last_seq. If the since is negative use self.last_seq.
        """
        if since < 0:
            since = self.last_seq
        data = self.get('/%s/_changes/?since=%s' % (self.name, since))
        self.last_seq = data['last_seq']
        return data

    def changesWithFilter(self, filter, limit=1000, since=-1):
        """
        Get the changes since sequence number. Store the last sequence value to
        self.last_seq. If the since is negative use self.last_seq.
        """
        if since < 0:
            since = self.last_seq
        data = self.get('/%s/_changes?limit=%s&since=%s&filter=%s' % (self.name, limit, since, filter))
        self.last_seq = data['last_seq']
        return data

    def purge(self, data):
        return self.post('/%s/_purge' % self.name, data)

    def loadView(self, design, view, options=None, keys=None):
        """
        Load a view by getting, for example:
        http://localhost:5984/tester/_view/viewtest/age_name?count=10&group=true

        The following URL query arguments are allowed:

        GET
                key=keyvalue
                startkey=keyvalue
                startkey_docid=docid
                endkey=keyvalue
                endkey_docid=docid
                limit=max rows to return
                stale=ok
                descending=true
                skip=number of rows to skip
                group=true Version 0.8.0 and forward
                group_level=int
                reduce=false Trunk only (0.9)
                include_docs=true Trunk only (0.9)
        POST
                {"keys": ["key1", "key2", ...]} Trunk only (0.9)

        more info: http://wiki.apache.org/couchdb/HTTP_view_API
        """
        options = options or {}
        keys = keys or []
        encodedOptions = {}
        for k, v in viewitems(options):
            # We can't encode the stale option, as it will be converted to '"ok"'
            # which couch barfs on.
            if k == "stale":
                encodedOptions[k] = v
            else:
                encodedOptions[k] = self.encode(v)

        if keys:
            if encodedOptions:
                data = urllib.parse.urlencode(encodedOptions)
                retval = self.post('/%s/_design/%s/_view/%s?%s' % \
                                   (self.name, design, view, data), {'keys': keys})
            else:
                retval = self.post('/%s/_design/%s/_view/%s' % \
                                   (self.name, design, view), {'keys': keys})
        else:
            retval = self.get('/%s/_design/%s/_view/%s' % \
                              (self.name, design, view), encodedOptions)
        if 'error' in retval:
            raise RuntimeError("Error in CouchDB: viewError '%s' reason '%s'" % \
                               (retval['error'], retval['reason']))
        else:
            return retval

    def loadList(self, design, list, view, options=None, keys=None):
        """
        Load data from a list function. This returns data that hasn't been
        decoded, since a list can return data in any format. It is expected that
        the caller of this function knows what data is being returned and how to
        deal with it appropriately.
        """
        options = options or {}
        keys = keys or []
        encodedOptions = {}
        for k, v in viewitems(options):
            encodedOptions[k] = self.encode(v)

        if keys:
            if encodedOptions:
                data = urllib.parse.urlencode(encodedOptions)
                retval = self.post('/%s/_design/%s/_list/%s/%s?%s' % \
                                   (self.name, design, list, view, data), {'keys': keys},
                                   decode=PY3_STR_DECODER)
            else:
                retval = self.post('/%s/_design/%s/_list/%s/%s' % \
                                   (self.name, design, list, view), {'keys': keys},
                                   decode=PY3_STR_DECODER)
        else:
            retval = self.get('/%s/_design/%s/_list/%s/%s' % \
                              (self.name, design, list, view), encodedOptions,
                              decode=PY3_STR_DECODER)

        return retval

    def getDoc(self, docName):
        """
        Return a single document from the database.
        """
        try:
            return self.get('/%s/%s' % (self.name, docName))
        except CouchError as e:
            # if empty dict, then doc does not exist in the db
            if getattr(e, "data", None) == {}:
                return {}
            self.checkForCouchError(getattr(e, "status", None), getattr(e, "reason", None))

    def allDocs(self, options=None, keys=None):
        """
        Return all the documents in the database
        options is a dict type parameter which can be passed to _all_docs
        id {'startkey': 'a', 'limit':2, 'include_docs': true}
        keys is the list of key (ids) for doc to be returned
        """
        options = options or {}
        keys = keys or []
        encodedOptions = {}
        for k, v in viewitems(options):
            encodedOptions[k] = self.encode(v)

        if keys:
            if encodedOptions:
                data = urllib.parse.urlencode(encodedOptions)
                return self.post('/%s/_all_docs?%s' % (self.name, data),
                                 {'keys': keys})
            else:
                return self.post('/%s/_all_docs' % self.name,
                                 {'keys': keys})
        else:
            return self.get('/%s/_all_docs' % self.name, encodedOptions)

    def info(self):
        """
        Return information about the databaes (size, number of documents etc).
        """
        return self.get('/%s/' % self.name)

    def addAttachment(self, id, rev, value, name=None, contentType=None, checksum=None, add_checksum=False):
        """
        Add an attachment stored in value to a document identified by id at revision rev.
        If specified the attachement will be uploaded as name, other wise the attachment is
        named "attachment".

        If not set CouchDB will try to determine contentType and default to text/plain.

        If checksum is specified pass this to CouchDB, it will refuse if the MD5 checksum
        doesn't match the one provided. If add_checksum is True calculate the checksum of
        the attachment and pass that into CouchDB for validation. The checksum should be the
        base64 encoded binary md5 (as returned by hashlib.md5().digest())
        """
        if name is None:
            name = "attachment"
        req_headers = {}

        if add_checksum:
            # calculate base64 encoded MD5
            keyhash = hashlib.md5()
            value_str = str(value) if not isinstance(value, (newstr, newbytes)) else value
            keyhash.update(encodeUnicodeToBytes(value_str))
            content_md5 = base64.b64encode(keyhash.digest())
            req_headers['Content-MD5'] = decodeBytesToUnicode(content_md5) if PY3 else content_md5
        elif checksum:
            req_headers['Content-MD5'] = decodeBytesToUnicode(checksum) if PY3 else checksum
        return self.put('/%s/%s/%s?rev=%s' % (self.name, id, name, rev),
                        value, encode=False,
                        contentType=contentType,
                        incoming_headers=req_headers)

    def getAttachment(self, id, name="attachment"):
        """
        _getAttachment_

        Retrieve an attachment for a couch document.
        """
        url = "/%s/%s/%s" % (self.name, id, name)
        attachment = self.get(url, None, encode=False, decode=PY3_STR_DECODER)

        # there has to be a better way to do this but if we're not de-jsoning
        # the return values, then this is all I can do for error checking,
        # right?
        # TODO: MAKE BETTER ERROR HANDLING
        if (attachment.find('{"error":"not_found","reason":"deleted"}') != -1):
            raise RuntimeError("File not found, deleted")
        if id == "nonexistantid":
            print(attachment)
        return attachment

    def bulkDeleteByIDs(self, ids):
        """
        delete bulk documents
        """
        # do the safety check other wise it will delete whole db.
        if not isinstance(ids, list):
            raise RuntimeError("Bulk delete requires a list of ids, wrong data type")
        if not ids:
            return None

        docs = self.allDocs(keys=ids)['rows']
        for j in docs:
            doc = {}
            if "id" not in j:
                print("Document not found: %s" % j)
                continue
            doc["_id"] = j['id']
            doc["_rev"] = j['value']['rev']
            self.queueDelete(doc)
        return self.commit()


class RotatingDatabase(Database):
    """
    A rotating database is actually multiple databases:
      - one active database (self)
      - N inactive databases (waiting to be removed)
      - one archive database
      - one configuration/seed database

    The active database is the one which serves current requests. It is active
    for a certain time window and then archived and marked as inactive.

    Inactive databases no longer recieve queries, although are still available
    on the server. They are queued up for deletion. This allows you to have a
    system where active databases are rotated daily and are kept in the server
    for a week. Inactive databases have a document in them defined as:
        {
          '_id': 'inactive',
          'archived_at': TIMESTAMP,     # added when archived
          'expires_at': TIMESTAMP+delta # added when archived
        }
    which is used to persist state across instatiations of the class.

    The archive database stores the results of views on the active databases
    once they are rotated out of service.

    The configuration/seed database holds the following information:
        * names of known inactive databases
        * name of current active database
        * name of archive database
        * design documents needed to seed new databases

    Once rotated the current active database is made inactive, a new active
    database created, views are copied to the archive database as necessary and
    the inactive databases queued for removal.
    """

    def __init__(self, dbname='database', url='http://localhost:5984',
                 size=1000, archivename=None, seedname=None,
                 timing=None, views=None):
        """
        dbaname:     base name for databases, active databases will have
                     timestamp appended
        url:         url of the CouchDB server
        size:        how big the data queue can get
        archivename: database to archive view results to, default is
                     dbname_archive
        seedname:    database where seed views and configuration/state are held
                     default is $dbname_seedcfg
        timing:      a dict containing two timedeltas 'archive' and 'expire',
                     if not present assume the database will br rotated by
                     external code
        views:       a list of views (design/name) to archive. The assumption
                     is that these views have been loaded into the seed
                     database via couchapp or someother process.
        """
        views = views or []
        # Store the base database name
        self.basename = dbname

        # Since we're going to be making databases hold onto a server
        self.server = CouchServer(url)

        # self is the "active" database
        Database.__init__(self, self._get_new_name(), url, size)
        # forcibly make sure I exist
        self.server.connectDatabase(self.name)

        # Set up the databases for the seed
        if not seedname:
            seedname = '%s_seedcfg' % (self.basename)
        self.seed_db = self.server.connectDatabase(seedname, url, size)

        # TODO: load a rotating DB from the seed db

        # TODO: Maybe call self._rotate() here?

        self.timing = timing

        self.archive_config = {}
        self.archive_db = None
        self.views = []
        if views:
            # If views isn't set in the constructor theres nothing to archive
            if not archivename:
                archivename = '%s_archive' % (self.basename)
            # TODO: check that the views listed exist in the seed
            # TODO: support passing in view options
            self.views = views
            self.archive_db = self.server.connectDatabase(archivename, url, size)
            self.archive_config['views'] = self.views
            self.archive_config['database'] = archivename
            self.archive_config['type'] = 'archive_config'
            self.archive_config['timing'] = str(self.timing)
            # copy views from the seed to the active db
            self._copy_views()
        if self.archive_config:
            # TODO: deal with multiple instances, load from doc?
            self.seed_db.commitOne(self.archive_config)

    def _get_new_name(self):
        return '%s_%s' % (self.basename, int(time.time()))

    def _copy_views(self):
        """
        Copy design documents from self.seed_db to the new active database.
        This means that all views in the design doc are copied, regardless of
        whether they are actually archived.
        """
        for design_to_copy in set(['_design/%s' % design.split('/')[0] for design in self.views]):
            design = self.seed_db.document(design_to_copy)
            del design['_rev']
            self.queue(design)
        self.commit()

    def _rotate(self):
        """
        Rotate the active database:
            1. create the new active database
            2. set self.name to the new database name
            3. write the inactive document to the old active database
            4. write the inactive document to the seed db
        """
        retiring_db = self.server.connectDatabase(self.name)
        # do the switcheroo
        new_active_db = self.server.connectDatabase(self._get_new_name())
        self.name = new_active_db.name
        self._copy_views()
        # "connect" to the old server, write inactive doc
        retiring_db.commitOne({'_id': 'inactive'}, timestamp=True)

        # record new inactive db to config
        # TODO: update function?

        state_doc = {'_id': retiring_db.name, 'rotate_state': 'inactive'}
        if not self.archive_config:
            # Not configured to archive anything, so skip inactive state
            # set the old db as archived instead
            state_doc['rotate_state'] = 'archived'
        self.seed_db.commitOne(state_doc, timestamp=True)

    def _archive(self):
        """
        Archive inactive databases
        """
        if self.archive_config:
            # TODO: This should be a worker thread/pool thingy so it's non-blocking
            for inactive_db in self.inactive_dbs():
                archiving_db = Database(inactive_db, self['host'])
                for view_to_archive in self.views:
                    # TODO: improve handling views and options here
                    design, view = view_to_archive.split('/')
                    for data in archiving_db.loadView(design, view, options={'group': True})['rows']:
                        self.archive_db.queue(data)
                self.archive_db.commit()
                # Now set the inactive view to archived
                db_state = self.seed_db.document(inactive_db)
                db_state['rotate_state'] = 'archived'
                self.seed_db.commit(db_state)

    def _expire(self):
        """
        Delete inactive databases that have expired, and remove state docs.
        """
        now = datetime.now()
        then = now - self.timing['expire']

        options = {'startkey': 0, 'endkey': int(time.mktime(then.timetuple()))}
        expired = self._find_dbs_in_state('archived', options)
        for db in expired:
            try:
                self.server.deleteDatabase(db['id'])
            except CouchNotFoundError:
                # if it's gone we don't care
                pass
            db_state = self.seed_db.document(db['id'])
            self.seed_db.queueDelete(db_state)
        self.seed_db.commit()

    def _create_design_doc(self):
        """Create a design doc with a view for the rotate state"""
        tempDesignDoc = {'views': {
                             'rotateState': {
                                 'map': "function(doc) {emit(doc.timestamp, doc.rotate_state, doc._id);}"
                                            },
                                   }
                         }
        self.seed_db.put('/%s/_design/TempDesignDoc' % self.seed_db.name, tempDesignDoc)

    def _find_dbs_in_state(self, state, options=None):
        """Creates a design document with a single (temporary) view in it"""
        options = options or {}
        if self.seed_db.documentExists("_design/TempDesignDoc"):
            logging.info("Skipping designDoc creation because it already exists!")
        else:
            self._create_design_doc()

        data = self.seed_db.loadView("TempDesignDoc", "rotateState", options=options)
        return data['rows']

    def inactive_dbs(self):
        """
        Return a list on inactive databases
        """
        return [doc['value'] for doc in self._find_dbs_in_state('inactive')]

    def archived_dbs(self):
        """
        Return a list of archived databases
        """
        return [doc['value'] for doc in self._find_dbs_in_state('archived')]

    def makeRequest(self, uri=None, data=None, type='GET', incoming_headers=None,
                    encode=True, decode=True, contentType=None,
                    cache=False, rotate=True):
        """
        Intercept the request, determine if I need to rotate, then carry out the
        request as normal.
        """
        incoming_headers = incoming_headers or {}
        if self.timing and rotate:

            # check to see whether I should rotate the database before processing the request
            db_age = datetime.fromtimestamp(float(self.name.split('_')[-1]))
            db_expires = db_age + self.timing['archive']
            if datetime.now() > db_expires:
                # save the current name for later
                old_db = self.name
                if self._queue:
                    # data I've got queued up should go to the old database
                    # can't call self.commit() due to recursion
                    uri = '/%s/_bulk_docs/' % self.name
                    data['docs'] = list(self._queue)
                    self.makeRequest(uri, data, 'POST', rotate=False)
                    self._reset_queue()
                self._rotate()  # make the new database
                # The uri passed in will be wrong, and the db may no longer exist if it has expired
                # so replacte the old name with the new
                uri.replace(old_db, self.name, 1)
        # write the data to the current database
        Database.makeRequest(self, uri, data, type, incoming_headers, encode, decode, contentType, cache)
        # now do some maintenance on the archived/expired databases
        self._archive()
        self._expire()


class CouchServer(CouchDBRequests):
    """
    An object representing the CouchDB server, use it to list, create, delete
    and connect to databases.

    More info http://wiki.apache.org/couchdb/HTTP_database_API
    """

    def __init__(self, dburl='http://localhost:5984', usePYCurl=True, ckey=None, cert=None, capath=None):
        """
        Set up a connection to the CouchDB server
        """
        check_server_url(dburl)
        CouchDBRequests.__init__(self, url=dburl, usePYCurl=usePYCurl, ckey=ckey, cert=cert, capath=capath)
        self.url = dburl
        self.ckey = ckey
        self.cert = cert

    def getCouchWelcome(self):
        """
        Retrieve CouchDB welcome information (which includes the version number)
        :return: a dictionary
        """
        return self.get('')

    def listDatabases(self):
        "List all the databases the server hosts"
        return self.get('/_all_dbs')

    def createDatabase(self, dbname, size=1000):
        """
        A database must be named with all lowercase characters (a-z),
        digits (0-9), or any of the _$()+-/ characters and must end with a slash
        in the URL.
        """
        check_name(dbname)

        self.put("/%s" % urllib.parse.quote_plus(dbname))
        # Pass the Database constructor the unquoted name - the constructor will
        # quote it for us.
        return Database(dbname=dbname, url=self.url, size=size, ckey=self.ckey, cert=self.cert)

    def deleteDatabase(self, dbname):
        """Delete a database from the server"""
        check_name(dbname)
        dbname = urllib.parse.quote_plus(dbname)
        if "cmsweb" in self.url:
            msg = f"You can't be serious that you want to delete a PRODUCTION database!!! "
            msg += f"At url: {self.url}, for database name: {dbname}. Bailing out!"
            raise RuntimeError(msg)
        return self.delete("/%s" % dbname)

    def connectDatabase(self, dbname='database', create=True, size=1000):
        """
        Return a Database instance, pointing to a database in the server. If the
        database doesn't exist create it if create is True.
        """
        check_name(dbname)
        if create and dbname not in self.listDatabases():
            return self.createDatabase(dbname)
        return Database(dbname=dbname, url=self.url, size=size, ckey=self.ckey, cert=self.cert)

    def replicate(self, source, destination, continuous=False,
                  create_target=False, cancel=False, doc_ids=False,
                  filter=False, query_params=False, sleepSecs=0, selector=False):
        """
        Trigger replication between source and destination. CouchDB options are
        defined in: https://docs.couchdb.org/en/3.1.2/api/server/common.html#replicate
        with further details in: https://docs.couchdb.org/en/stable/replication/replicator.html

        Source and destination need to be appropriately urlquoted after the port
        number. E.g. if you have a database with /'s in the name you need to
        convert them into %2F's.

        TODO: Improve source/destination handling - can't simply URL quote,
        though, would need to decompose the URL and rebuild it.

        :param source: string with the source url to replicate data from
        :param destination: string with the destination url to replicate data to
        :param continuous: boolean to perform a continuous replication or not
        :param create_target: boolean to create the target database, if non-existent
        :param cancel: boolean to stop a replication (but we better just delete the doc!)
        :param doc_ids: a list of specific doc ids that we would like to replicate
        :param filter: string with the name of the filter function to be used. Note that
                       this filter is expected to have been defined in the design doc.
        :param query_params: dictionary of parameters to pass over to the filter function
        :param sleepSecs: amount of seconds to sleep after the replication job is created
        :param selector: a new'ish feature for filter functions in Erlang
        :return: status of the replication creation
        """
        listDbs = self.listDatabases()
        if source not in listDbs:
            check_server_url(source)
        if destination not in listDbs:
            if create_target and not destination.startswith("http"):
                check_name(destination)
            else:
                check_server_url(destination)

        if not destination.startswith("http"):
            destination = '%s/%s' % (self.url, destination)
        if not source.startswith("http"):
            source = '%s/%s' % (self.url, source)
        data = {"source": source, "target": destination}
        # There must be a nicer way to do this, but I've not had coffee yet...
        if continuous: data["continuous"] = continuous
        if create_target: data["create_target"] = create_target
        if cancel: data["cancel"] = cancel
        if doc_ids: data["doc_ids"] = doc_ids
        if filter:
            data["filter"] = filter
            if query_params:
                data["query_params"] = query_params
        if selector: data["selector"] = selector

        resp = self.post('/_replicator', data)
        # Sleep required for CouchDB 3.x unit tests
        time.sleep(sleepSecs)
        return resp

    def status(self):
        """
        See what active tasks are running on the server.
        """
        return {'databases': self.listDatabases(),
                'server_stats': self.get('/_stats'),
                'active_tasks': self.get('/_active_tasks')}

    def __str__(self):
        """
        List all the databases the server has
        """
        return self.listDatabases().__str__()


# define some standard couch error classes
# from:
#  http://wiki.apache.org/couchdb/HTTP_status_list

class CouchError(Exception):
    "An error thrown by CouchDB"

    def __init__(self, reason, data, result, status=None):
        Exception.__init__(self)
        self.reason = reason
        self.data = data
        self.result = result
        self.type = "CouchError"
        self.status = status

    def __str__(self):
        """Stringify the error"""
        errorMsg = ""
        if self.type == "CouchError":
            errorMsg += "A NEW COUCHDB ERROR TYPE/STATUS HAS BEEN FOUND! "
            errorMsg += "UPDATE CMSCOUCH.PY IMPLEMENTATION WITH A NEW COUCH ERROR/STATUS! "
            errorMsg += f"Status: {self.status}\n"
        errorMsg += f"Error type: {self.type}, Status code: {self.status}, "
        errorMsg += f"Reason: {self.reason}, Data: {repr(self.data)}"
        return errorMsg


class CouchBadRequestError(CouchError):
    def __init__(self, reason, data, result, status):
        CouchError.__init__(self, reason, data, result, status)
        self.type = "CouchBadRequestError"


class CouchUnauthorisedError(CouchError):
    def __init__(self, reason, data, result, status):
        CouchError.__init__(self, reason, data, result, status)
        self.type = "CouchUnauthorisedError"


class CouchNotFoundError(CouchError):
    def __init__(self, reason, data, result, status):
        CouchError.__init__(self, reason, data, result, status)
        self.type = "CouchNotFoundError"


class CouchNotAllowedError(CouchError):
    def __init__(self, reason, data, result, status):
        CouchError.__init__(self, reason, data, result, status)
        self.type = "CouchNotAllowedError"

class CouchNotAcceptableError(CouchError):
    def __init__(self, reason, data, result, status):
        CouchError.__init__(self, reason, data, result, status)
        self.type = "CouchNotAcceptableError"

class CouchConflictError(CouchError):
    def __init__(self, reason, data, result, status):
        CouchError.__init__(self, reason, data, result, status)
        self.type = "CouchConflictError"


class CouchFeatureGone(CouchError):
    def __init__(self, reason, data, result, status):
        CouchError.__init__(self, reason, data, result, status)
        self.type = "CouchFeatureGone"


class CouchPreconditionFailedError(CouchError):
    def __init__(self, reason, data, result, status):
        CouchError.__init__(self, reason, data, result, status)
        self.type = "CouchPreconditionFailedError"


class CouchRequestTooLargeError(CouchError):
    def __init__(self, reason, data, result, status):
        # calculate the size of this JSON serialized object
        docSize = sys.getsizeof(json.dumps(data))
        errorMsg = f"Document has {docSize} bytes and it's too large to be accepted by CouchDB. "
        errorMsg += f"Check the CouchDB configuration to see the current value "
        errorMsg += f"under 'couchdb.max_document_size' (default is 8M bytes)."
        CouchError.__init__(self, reason, errorMsg, result, status)
        self.type = "CouchRequestTooLargeError"


class CouchExpectationFailedError(CouchError):
    def __init__(self, reason, data, result, status):
        CouchError.__init__(self, reason, data, result, status)
        self.type = "CouchExpectationFailedError"

class CouchRequestedRangeNotSatisfiableError(CouchError):
    def __init__(self, reason, data, result, status):
        CouchError.__init__(self, reason, data, result, status)
        self.type = "CouchRequestedRangeNotSatisfiableError"


class CouchInternalServerError(CouchError):
    def __init__(self, reason, data, result, status):
        CouchError.__init__(self, reason, data, result, status)
        self.type = "CouchInternalServerError"


class CouchForbidden(CouchError):
    def __init__(self, reason, data, result, status):
        CouchError.__init__(self, reason, data, result, status)
        self.type = "CouchForbidden"


class CouchMonitor(object):
    def __init__(self, couchURL):
        if isinstance(couchURL, CouchServer):
            self.couchServer = couchURL
        else:
            self.couchServer = CouchServer(couchURL)

        self.replicatorDB = self.couchServer.connectDatabase('_replicator', False)

        # use the CouchDB version to decide which APIs and schema is available
        couchInfo = self.couchServer.getCouchWelcome()
        self.couchVersion = couchInfo.get("version")

    def deleteReplicatorDocs(self, source=None, target=None, repDocs=None):
        if repDocs is None:
            repDocs = self.replicatorDB.allDocs(options={'include_docs': True})['rows']

        filteredDocs = self._filterReplicationDocs(repDocs, source, target)
        if not filteredDocs:
            return
        for doc in filteredDocs:
            self.replicatorDB.queueDelete(doc)
        return self.replicatorDB.commit()

    def _filterReplicationDocs(self, repDocs, source, target):
        filteredDocs = []
        for j in repDocs:
            if '_design' not in j['id']:
                if (source is None and target is None) or \
                        (j['doc']['source'] == source and j['doc']['target'] == target):
                    doc = {}
                    doc["_id"] = j['id']
                    doc["_rev"] = j['value']['rev']
                    filteredDocs.append(doc)
        return filteredDocs

    def getActiveTasks(self):
        """
        Return all the active tasks in Couch (compaction, replication, indexing, etc)
        :return: a list with the current active tasks

        For further information:
            https://docs.couchdb.org/en/3.1.2/api/server/common.html#active-tasks
        """
        return self.couchServer.get("/_active_tasks")

    def getSchedulerJobs(self):
        """
        Return all replication jobs created either via _replicate or _replicator dbs.
        It does not include replications that have either completed or failed.
        :return: a list with the current replication jobs

        For further information:
            https://docs.couchdb.org/en/3.1.2/api/server/common.html#api-server-scheduler-jobs
        """
        resp = []
        data = self.couchServer.get("/_scheduler/jobs")
        return data.get("jobs", resp)

    def getSchedulerDocs(self):
        """
        Return all replication documents and their states, even if they have completed or
        failed.
        :return: a list with the current replication docs

        Replication states can be found at:
            https://docs.couchdb.org/en/3.1.2/replication/replicator.html#replicator-states
        For further information:
            https://docs.couchdb.org/en/3.1.2/api/server/common.html#api-server-scheduler-docs
        """
        # NOTE: if there are no docs, this call can give a response like:
        # {"error":"not_found","reason":"Database does not exist."}
        resp = []
        try:
            data = self.couchServer.get("/_scheduler/docs")
        except CouchNotFoundError as exc:
            logging.warning("/_scheduler/docs API returned: %s", getattr(exc, "result", ""))
            return resp
        return data.get("docs", resp)

    def couchReplicationStatus(self):
        """
        check couchdb replication status with compatible output of checkCouchReplications

        :return: a list of dictionaries with the status of the replications and an
            error message
        """
        # Do not import checkStatus at the beginning of the CMSCouch module
        # to avoid dependecy conflicts in the CMS Sandbox at the WN level
        from WMCore.Database.CouchMonitoring import checkStatus

        output = []
        sdict = checkStatus(kind='scheduler')
        rdict = checkStatus(kind='replicator')
        method = 'scheduler+replicator'
        # update sdict only with entries from replicator dict which are not present in scheduler
        for key, val in rdict['current_status'].items():
            if key not in sdict['current_status']:
                sdict['current_status'][key] = val
        stateFailures = ['error', 'failed']
        for rid, record in sdict['current_status'].items():
            if record['state'] in stateFailures:
                status['state'] = 'error'
                source = sanitizeURL(record['source'])
                target = sanitizeURL(record['target'])
                error = record['error']
                history = pformat(record['history'])
                msg = f"Replication from {source} to {target} for document {rid} is in a bad state: {error}; "
                msg += f"History: {history}"
                status = {'name': 'CouchServer', 'status': 'error', 'error_message': msg, 'method': method}
                output.append(status)

        # if our replication is fine we should check that it is not in a stale phase
        activeTasks = self.getActiveTasks()
        activeTasks = [task for task in activeTasks if task["type"].lower() == "replication"]
        resp = self.checkReplicationState()
        for replTask in activeTasks:
            if self.isReplicationStale(replTask):
                source = sanitizeURL(replTask['source'])['url']
                target = sanitizeURL(replTask['target'])['url']
                msg = f"Replication from {source} to {target} is stale and it's last"
                msg += f"update time was at: {replTask.get('updated_on')}"
                resp['status'] = 'error'
                resp['error_message'] += msg
                resp['method'] = 'stale phase'
                resp['name'] = 'CouchServer'
                output.append(resp)
        # check if we did not record any replication status, then add the ok status
        if len(output) == 0:
            status = {'name': 'CouchServer', 'status': 'ok', 'error_message': ''}
            output.append(status)
        return output

    def checkCouchReplications(self, replicationsList):
        """
        Check whether the list of expected replications exist in CouchDB
        and also check their status.

        :param replicationsList: a list of dictionary with the replication
            document setup.
        :return: a list of dictionaries with the status of the replications and an
            error message
        """
        output = []
        method = 'comparison of replications docs vs active tasks'
        activeTasks = self.getActiveTasks()
        # filter out any task that is not a database replication
        activeTasks = [task for task in activeTasks if task["type"].lower() == "replication"]

        if len(replicationsList) != len(activeTasks):
            msg = f"Expected to have {len(replicationsList)} replication tasks, "
            msg += f"but only {len(activeTasks)} in CouchDB. "
            msg += f"Current replications are: {activeTasks}"
            status = {'name': 'CouchServer', 'status': 'error', 'error_message': msg, 'method': method}
            output.append(status)

        resp = self.checkReplicationState()
        if resp['status'] != 'ok':
            output.append(resp)

        # finally, check if replications are being updated in a timely fashion
        for replTask in activeTasks:
            if not self.isReplicationOK(replTask):
                source = sanitizeURL(replTask['source'])['url']
                target = sanitizeURL(replTask['target'])['url']
                msg = f"Replication from {source} to {target} is stale and it's last"
                msg += f"update time was at: {replTask.get('updated_on')}"
                resp['status'] = 'error'
                resp['error_message'] += msg
                resp['method'] = method
                resp['name'] = 'CouchServer'
                output.append(resp)

        # check if we did not record any replication status, then add the ok status
        if len(output) == 0:
            status = {'name': 'CouchServer', 'status': 'ok', 'error_message': ''}
            output.append(status)
        return output

    def checkReplicationState(self):
        """
        Check the state of the existent replication tasks.
        NOTE that this can't be done for CouchDB 1.6, since there is
        replication state.

        :return: a dictionary with the status of the replications and an
                 error message
        """
        resp = {'status': 'ok', 'error_message': ""}
        if self.couchVersion == "1.6.1":
            return resp

        for replDoc in self.getSchedulerDocs():
            if replDoc['state'].lower() not in ["pending", "running"]:
                source = sanitizeURL(replDoc['source'])['url']
                target = sanitizeURL(replDoc['target'])['url']
                msg = f"Replication from {source} to {target} is in a bad state: {replDoc.get('state')}; "
                resp['status'] = "error"
                resp['error_message'] += msg
        return resp

    def isReplicationOK(self, replInfo):
        """
        Ensure that the replication document is up-to-date as a
        function of the checkpoint interval.

        :param replInfo: dictionary with the replication information
        :return: True if replication is working fine, otherwise False
        """
        maxUpdateInterval = replInfo['checkpoint_interval'] / 1000
        lastUpdate = replInfo["updated_on"]

        if lastUpdate + maxUpdateInterval > int(time.time()):
            # then it has been recently updated
            return True
        return False

    def isReplicationStale(self, replInfo, niter=10):
        """
        Ensure that the replication document is up-to-date as a
        function of the checkpoint interval.

        :param replInfo: dictionary with the replication information
        :param niter: number of iteration for checkpoint interval
        :return: True if replication is working fine, otherwise False
        """
        maxUpdateInterval = niter * replInfo['checkpoint_interval'] / 1000
        lastUpdate = replInfo["updated_on"]

        if lastUpdate + maxUpdateInterval > int(time.time()):
            # then it has been recently updated and it means replication is not stale
            return False
        return True



#!/usr/bin/env python
"""
_Update_

Oracle implementation of Files.Update

Created on Feb 25, 2013

@author: dballest
"""

from WMCore.WMBS.MySQL.Files.Update import Update as MySQLUpdate

class Update(MySQLUpdate):
    pass



"""
__JobUpdater__

Check and calculate changes in job information and
updates the records of pending jobs in the batch
system

Created on Apr 16, 2013

@author: dballest
"""

import logging
import threading

from WMComponent.JobUpdater.JobUpdaterPoller import JobUpdaterPoller

from WMCore.Agent.Harness import Harness

class JobUpdater(Harness):
    """
    Component class for the JobUpdater module,
    runs a single worker, the JobUpdaterPoller
    """

    def __init__(self, config):
        """
        __init__

        Initialize the Harness
        """
        Harness.__init__(self, config)

    def preInitialization(self):
        """
        _preInitialization_

        Sets up the worker thread
        """
        logging.info("JobTracker.preInitialization")

        myThread = threading.currentThread()

        pollInterval = self.config.JobUpdater.JobUpdaterPoller.pollInterval
        logging.info("Setting poll interval to %s seconds" % pollInterval)
        myThread.workerThreadManager.addWorker(JobUpdaterPoller(self.config), pollInterval)

        return



"""
_WMWorkloadTools_

Define some generic tools used by the StdSpecs and WMWorkload
to validate arguments that modify a WMWorkload and/or WMTask.

Created on Jun 13, 2013

@author: dballest
"""
from builtins import str, bytes
from future.utils import viewitems, viewvalues

import json
import re
import inspect

from Utils.Utilities import makeList
from WMCore.DataStructs.LumiList import LumiList
from WMCore.WMSpec.WMSpecErrors import WMSpecFactoryException

def checkMemCore(paramInfo, minValue=1):
    """
    Spec validation function for the Memory/Multicore parameters
    :param memInfo: memory provided by the user (can be an int/float/dict)
    :param minValue: base value to be used to validate user's values
    :return: True if validation is successful, raise an exception otherwise
    """
    try:
        if isinstance(paramInfo, dict):
            for value in viewvalues(paramInfo):
                assert value >= minValue
        else:
            assert paramInfo >= minValue
    except Exception:
        return False
    return True


def checkEventStreams(streamsInfo):
    """
    Spec validation function for the EventStreams parameter
    :param streamsInfo: event streams provided by the user (can be either int or a dict)
    :return: True if validation is successful, raise an exception otherwise
    """
    return checkMemCore(streamsInfo, 0)


def checkTimePerEvent(tpEvtInfo):
    """
    Spec validation function for the TimePerEvent parameter
    :param tpEvtInfo: time per event provided by the user (can be either float or a dict)
    :return: True if validation is successful, raise an exception otherwise
    """
    return checkMemCore(tpEvtInfo, 0)


def makeLumiList(lumiDict):
    try:
        if isinstance(lumiDict, (str, bytes)):
            lumiDict = json.loads(lumiDict)
        ll = LumiList(compactList=lumiDict)
        return ll.getCompactList()
    except Exception:
        raise WMSpecFactoryException("Could not parse LumiList, %s: %s" % (type(lumiDict), lumiDict))


def parsePileupConfig(mcPileup, dataPileup):
    """
    _parsePileupConfig_

    If the pileup config is defined as MCPileup and DataPileup
    then make sure we get the usual dictionary as
    PileupConfig : {'mc': ['/mc_pd/procds/tier'], 'data': ['/data_pd/procds/tier']}
    """
    pileUpConfig = {}
    if mcPileup is not None:
        pileUpConfig['mc'] = [mcPileup]
    if dataPileup is not None:
        pileUpConfig['data'] = [dataPileup]
    return pileUpConfig


def _validateArgument(argument, value, argumentDefinition):
    """
    Validate a single argument against its definition in the spec
    """
    validNull = argumentDefinition["null"]
    if not validNull and value is None:
        raise WMSpecFactoryException("Argument '%s' cannot be None" % argument)
    elif value is None:
        return value

    # is it a built-in type (int, str, bool, etc.) or a custom function
    expected_type = argumentDefinition["type"]
    try:
        if expected_type == str and not isinstance(value, str):
            raise TypeError(f"Argument '{argument}' with value {value} is not a string")
        else:
            # for other data types (or custom functions), just try to cast it to the expected type
            value = expected_type(value)
    except (ValueError, TypeError) as e:
        msg = "Argument '%s' with value %r, has an incorrect data type: " % (argument, value)
        msg += "%s. It must be convertible by %s" % (type(value), expected_type.__name__)
        if str(e):
            msg += f" (Error: {e})"
        raise WMSpecFactoryException(msg) from None
    except Exception as e:
        msg = f"Generic exception raised during argument validation for argument: {argument} "
        msg += "with value: %r. Error details: %s" % (value, str(e))
        raise WMSpecFactoryException(msg) from None

    _validateArgFunction(argument, value, argumentDefinition["validate"])
    return value


def _validateArgumentDict(argument, argValue, argumentDefinition):
    """
    Validate only the basic structure of dict arguments, we anyways
    don't have the definition of the internal arguments.
    """
    # make sure we're not going to cast a dict to string and let that unnoticed
    if isinstance(argumentDefinition["type"], type(dict)) and not isinstance(argValue, dict):
        msg = "Argument '%s' with value %r, has an incorrect data type: " % (argument, argValue)
        msg += "%s. It must be %s" % (type(argValue), argumentDefinition["type"])
        raise WMSpecFactoryException(msg)

    # still an exception, make sure it has the correct format
    if argument == "LumiList":
        argValue = argumentDefinition["type"](argValue)

    _validateArgFunction(argument, argValue, argumentDefinition["validate"])
    return argValue


def _validateArgFunction(argument, value, valFunction):
    """
    Perform the validation function as in the argument definition
    """
    if valFunction:
        try:
            if not valFunction(value):
                msg = "Argument '%s' with value %r, doesn't pass the validate function." % (argument, value)
                msg += "\nIt's definition is:\n%s" % inspect.getsource(valFunction)
                raise WMSpecFactoryException(msg)
        except WMSpecFactoryException:
            # just re-raise it to keep the error message clear
            raise
        except Exception as ex:
            # Some validation functions (e.g. Lexicon) will raise errors instead of returning False
            raise WMSpecFactoryException(str(ex))
    return


def _validateArgumentOptions(arguments, argumentDefinition, optionKey=None):
    """
    Check whether create or assign mandatory parameters were properly
    set in the request schema.
    """
    for arg, argDef in viewitems(argumentDefinition):
        optional = argDef.get(optionKey, True)
        if not optional and arg not in arguments:
            msg = "Argument '{}' is mandatory and must be provided by the user.".format(arg)
            msg += " Its definition is: {}".format(argDef)
            raise WMSpecFactoryException(msg)
        # specific case when user GUI returns empty string for optional arguments
        elif arg not in arguments:
            continue
        elif isinstance(arguments[arg], dict):
            arguments[arg] = _validateArgumentDict(arg, arguments[arg], argDef)
        else:
            arguments[arg] = _validateArgument(arg, arguments[arg], argDef)
    return


def validateInputDatasSetAndParentFlag(arguments):
    """
    Check if the InputDataset value provided corresponds to an actual dataset in DBS.
    If parent flag is provided, then check whether the input dataset has a parent.
    the InputDataset existence in DBS and its parent, if needed.
    """
    inputdataset = _getChainKey(arguments, "InputDataset")
    mcpileup = _getChainKey(arguments, "MCPileup")
    datapileup = _getChainKey(arguments, "DataPileup")
    includeParents = _getChainKey(arguments, "IncludeParents")
    # TODO: this replace can be removed in one year from now, thus March 2022
    dbsURL = arguments.get("DbsUrl")
    if dbsURL:
        dbsURL = dbsURL.replace("cmsweb.cern.ch", "cmsweb-prod.cern.ch")
        dbsURL = dbsURL.rstrip("/")

    if includeParents and not inputdataset:
        msg = "IncludeParents flag is True but InputDataset value has not been provided"
        raise WMSpecFactoryException(msg)

    if dbsURL and inputdataset or mcpileup or datapileup:
        # import DBS3Reader here, since Runtime code import this module and worker
        # node doesn't have dbs3 client
        from WMCore.Services.DBS.DBS3Reader import DBS3Reader
        from WMCore.Services.DBS.DBSErrors import DBSReaderError
        dbsInst = DBS3Reader(dbsURL)

        try:
            _datasetExists(dbsInst, inputdataset)
            _datasetExists(dbsInst, mcpileup)
            _datasetExists(dbsInst, datapileup)
        except DBSReaderError as ex:
            # we need to Wrap the exception to WMSpecFactoryException to be caught in reqmgr validation
            raise WMSpecFactoryException(str(ex))

        if includeParents:
            try:
                result = dbsInst.listDatasetParents(inputdataset)
                if len(result) == 0:
                    msg = "IncludeParents flag is True but the input dataset %s has no parents" % inputdataset
                    raise DBSReaderError(msg)
            except DBSReaderError as ex:
                raise WMSpecFactoryException(str(ex))

    return


def _datasetExists(dbsInst, inputData):
    """
    __datasetExists_

    Check if dataset exists in DBS. Exception is raised in case it does not exist.
    """
    if inputData is None:
        return
    dbsInst.checkDatasetPath(inputData)
    return


def _getChainKey(arguments, keyName):
    """
    Given a request arguments dictionary and a key name, properly returns its
    value regardless of the request type.
    """
    if "TaskChain" in arguments:
        value = arguments['Task1'].get(keyName)
    elif "StepChain" in arguments:
        value = arguments['Step1'].get(keyName)
    else:
        value = arguments.get(keyName)
    return value


def _validateMoveSubscription(subType, sites):
    """
    Move subscriptions are only allowed to T0 or T1s, see #7760
    """
    invalidSites = [site for site in sites if re.match("^T[2-3]", site)]
    if invalidSites:
        msg = "Move subscription (%s) not allowed to T2/T3 sites: %s" % (subType, invalidSites)
        raise WMSpecFactoryException(msg)


def validateSiteLists(arguments):
    whiteList = arguments.get("SiteWhitelist", [])
    blackList = arguments.get("SiteBlacklist", [])
    whiteList = makeList(whiteList)
    blackList = makeList(blackList)
    res = (set(whiteList) & set(blackList))
    if len(res):
        msg = "Validation failed: The same site cannot be white and blacklisted: %s" % list(res)
        raise WMSpecFactoryException(msg)
    # store the properly formatted values (list instead of string)
    arguments["SiteWhitelist"] = whiteList
    arguments["SiteBlacklist"] = blackList
    return


def validateArgumentsCreate(arguments, argumentDefinition, checkInputDset=True):
    """
    _validateArguments_

    Validate a set of arguments - for spec creation - against their
    definition in StdBase.getWorkloadCreateArgs.

    When validating Step/Task dictionary, checkInputDset should be usually
    false since the input dataset validation already happened at top level.

    It returns an error message if the validation went wrong,
    otherwise returns None.
    """
    validateUnknownArgs(arguments, argumentDefinition)
    _validateArgumentOptions(arguments, argumentDefinition, "optional")
    if checkInputDset:
        validateInputDatasSetAndParentFlag(arguments)
    return


def validateArgumentsUpdate(arguments, argumentDefinition, optionKey="assign_optional"):
    """
    _validateArgumentsUpdate_

    Validate a set of arguments - for spec assignment/update - against
    their definition in StdBase.getWorkloadAssignArgs.

    It returns an error message if the validation went wrong,
    otherwise returns None.
    """
    validateUnknownArgs(arguments, argumentDefinition)
    _validateArgumentOptions(arguments, argumentDefinition, optionKey)
    validateSiteLists(arguments)

    return

def validateUnknownArgs(arguments, argumentDefinition):
    """
    Make sure user is sending only arguments that are known by
    StdBase.getWorkloadCreateArgs, otherwise fail spec creation.

    It returns an error message if the validation went wrong,
    otherwise returns None.
    """
    unknownArgs = set(arguments) - set(argumentDefinition.keys())
    if unknownArgs:
        # now onto the exceptions...
        if arguments.get("RequestType") == "ReReco":
            unknownArgs = unknownArgs - set([x for x in unknownArgs if x.startswith("Skim")])
        elif arguments.get("RequestType") == "StepChain":
            unknownArgs = unknownArgs - set([x for x in unknownArgs if x.startswith("Step")])
        elif arguments.get("RequestType") == "TaskChain":
            unknownArgs = unknownArgs - set([x for x in unknownArgs if x.startswith("Task")])
        elif arguments.get("RequestType") == "Resubmission":
            # oh well, then we have to skip all possible obscure arguments
            unknownArgs = unknownArgs - set([x for x in unknownArgs if x.startswith("Skim")])
            unknownArgs = unknownArgs - set([x for x in unknownArgs if x.startswith("Step")])
            unknownArgs = unknownArgs - set([x for x in unknownArgs if x.startswith("Task")])

        if unknownArgs:
            msg = "There are unknown/unsupported arguments in your request spec: %s" % list(unknownArgs)
            raise WMSpecFactoryException(msg)
    return


def setArgumentsWithDefault(arguments, argumentDefinition):
    """
    Set arguments not provided by the user with a default spec value
    """
    for argument in argumentDefinition:
        if argument not in arguments:
            arguments[argument] = argumentDefinition[argument]["default"]

    # set the Campaign default value to the same as AcquisitionEra if Campaign is not specified
    if "Campaign" in argumentDefinition and not arguments.get("Campaign"):
        if "AcquisitionEra" in arguments and isinstance(arguments["AcquisitionEra"], (str, bytes)):
            arguments["Campaign"] = arguments["AcquisitionEra"]

    return

def setAssignArgumentsWithDefault(arguments, argumentDefinition):
    """
    Set arguments not provided by the user with a default assign spec value,
    unless the default value is None (read don't set default).
    """
    for argument in argumentDefinition:
        if argument not in arguments and argumentDefinition[argument]["default"] is not None:
            arguments[argument] = argumentDefinition[argument]["default"]

    return

def loadSpecClassByType(specType):
    factoryName = "%sWorkloadFactory" % specType
    mod = __import__("WMCore.WMSpec.StdSpecs.%s" % specType,
                     globals(), locals(), [factoryName])
    specClass = getattr(mod, factoryName)

    return specClass


def loadSpecByType(specType):
    specClass = loadSpecClassByType(specType)
    return specClass()


def checkDBSURL(url):
    # import DBS3Reader here, since Runtime code import this module and worker node doesn't have dbs3 client
    from WMCore.Services.DBS.DBS3Reader import DBS3Reader
    return DBS3Reader(url).checkDBSServer()
